#cython: boundscheck=False, wraparound=False, initializedcheck=False

cimport numpy as np
import numpy as np
import cython

from finesse.cymath cimport complex_t
from finesse.cymath.complex cimport conj, creal, cimag, cexp, carg, cabs
from finesse.cymath.complex cimport crotate, cnorm
from finesse.cymath.math cimport sqrt, sin, cos, float_eq
from finesse.cymath.homs cimport in_mask, field_index

from finesse.knm cimport KnmMatrix
from finesse.simulations.basematrix cimport CarrierSignalMatrixSimulation
from finesse.frequency cimport frequency_info_t
from finesse.detectors.workspace cimport (
    DetectorWorkspace,
    MaskedDetectorWorkspace,
    OutputFuncWrapper,
)
from finesse.element cimport BaseCValues

cimport finesse.constants

from libc.math cimport isnan
from libc.stdlib cimport calloc, realloc, free

ctypedef (double*,) ptr_tuple_1
ctypedef (double*, double*) ptr_tuple_2
ctypedef (double*, double*, double*) ptr_tuple_3
ctypedef (double*, double*, double*, double*) ptr_tuple_4
ctypedef (double*, double*, double*, double*, double*) ptr_tuple_5
ctypedef (double*, double*, double*, double*, double*, double*) ptr_tuple_6
ctypedef (double*, double*, double*, double*, double*, double*, double*) ptr_tuple_7
ctypedef (double*, double*, double*, double*, double*, double*, double*, double*) ptr_tuple_8
ctypedef (double*, double*, double*, double*, double*, double*, double*, double*, double*) ptr_tuple_9
ctypedef (double*, double*, double*, double*, double*, double*, double*, double*, double*, double*) ptr_tuple_10
ctypedef (double*, double*, double*, double*, double*, double*, double*, double*, double*, double*, double*) ptr_tuple_11

cdef extern from "constants.h":
    double complex COMPLEX_0

cdef:
    double NAN = 0.0/0.0


### PD0 workspace & output funcs ###

cdef class PD0Workspace(MaskedDetectorWorkspace):
    cdef public:
        int dc_node_id
        Py_ssize_t rhs_index
        Py_ssize_t size
        const complex_t[:,::1] K
        np.ndarray tmp


pd0_DC_output = OutputFuncWrapper.make_from_ptr(c_pd0_DC_output)
cdef c_pd0_DC_output(DetectorWorkspace dws):
    """This expects a `PD0Workspace` input. Fast computation of DC power at a node."""
    cdef:
        PD0Workspace ws = <PD0Workspace> dws
        Py_ssize_t i
        double tmp = 0

    for i in range(ws.rhs_index, ws.rhs_index + ws.size):
        tmp += cnorm(ws.sim.carrier.out_view[i])

    return tmp * ws.sim.model_data.EPSILON0_C * 0.5


pd0_DC_output_segmented = OutputFuncWrapper.make_from_ptr(c_pd0_DC_output_segmented)
cdef c_pd0_DC_output_segmented(DetectorWorkspace dws):
    """This expects a `PD0Workspace` input. Fast computation of DC power at a node
    with segmentation assumed in the photodiode. Where :math:`E` is the vector of
    complex fields, and :math:`K` is a Knm matrix describing the segementation of
    a detector, the output is:
    .. math::
        E^{\ast} \cdot (K \cdot E)
    """
    cdef:
        PD0Workspace ws = <PD0Workspace> dws
        Py_ssize_t i
        complex_t res = 0
        const complex_t[::1] E
        frequency_info_t *f1
        Py_ssize_t Nf = ws.sim.carrier.optical_frequencies.size
    for i in range(Nf):
        f1 = &ws.sim.carrier.optical_frequencies.frequency_info[i]
        rhs_idx = ws.sim.carrier.field_fast(ws.dc_node_id, f1.index, 0)
        E = ws.sim.carrier.out_view[rhs_idx:(rhs_idx + ws.sim.model_data.num_HOMs)]
        # segmented Knm multiply by the array of fields
        tmp = np.dot(ws.K, E)
        # inner product with conjugate
        res += np.vdot(tmp, E)
    # Only select the real part of this
    return res.real * ws.sim.model_data.EPSILON0_C * 0.5


pd0_DC_output_masked = OutputFuncWrapper.make_from_ptr(c_pd0_DC_output_masked)
cdef c_pd0_DC_output_masked(DetectorWorkspace dws):
    """This expects a `PD0Workspace` input. Fast computation of DC power at a node where
    a HOM mask has been applied to the PD."""
    cdef:
        PD0Workspace ws = <PD0Workspace> dws
        Py_ssize_t i, j, k
        Py_ssize_t freq_idx
        double tmp = 0

    for i in range(ws.sim.carrier.optical_frequencies.size):
        freq_idx = ws.rhs_index + i
        for j in range(ws.num_unmasked_HOMs):
            k = freq_idx + ws.unmasked_mode_indices[j]
            tmp += cnorm(ws.sim.carrier.out_view[k])

    return tmp * ws.sim.model_data.EPSILON0_C * 0.5


### PD1 workspace & output funcs ###

cdef class PD1Values(BaseCValues):
    def __init__(self):
        cdef ptr_tuple_2 ptr = (&self.f, &self.phase)
        cdef tuple params = ("f", "phase")
        self.setup(params, sizeof(ptr), <double**>&ptr)


cdef class PD1Workspace(MaskedDetectorWorkspace):
    def __cinit__(self, *args, **kwargs):
        self.num_pre_set_beats = 0
        self.pre_set_beats[0] = NULL
        self.pre_set_beats[1] = NULL

    def __init__(self, owner, sim, *, phase_offset=0, oinfo=None):
        super().__init__(owner, sim, PD1Values(), oinfo=oinfo)
        self.v = <PD1Values>self.values
        self.phase_offset = phase_offset

    cpdef update_beats(self):
        """
        Requires full initialisation and `self.update_parameter_values` calling
        before using
        """
        cdef:
            list beats = []
            frequency_info_t *f1
            frequency_info_t *f2
            Py_ssize_t i, j, n, prev_num_beats
            Py_ssize_t Nf = self.sim.carrier.optical_frequencies.size

        prev_num_beats = self.num_pre_set_beats

        for i in range(Nf):
            f1 = &self.sim.carrier.optical_frequencies.frequency_info[i]
            for j in range(Nf):
                f2 = &self.sim.carrier.optical_frequencies.frequency_info[j]
                if f1.f - f2.f == self.v.f:
                    beats.append((
                        self.sim.carrier.field_fast(self.dc_node_id, f1.index, 0),
                        self.sim.carrier.field_fast(self.dc_node_id, f2.index, 0),
                    ))

        self.num_pre_set_beats = len(beats)

        if self.num_pre_set_beats > 0:
            if prev_num_beats > 0:
                self.pre_set_beats[0] = <Py_ssize_t*>realloc(
                    self.pre_set_beats[0],
                    self.num_pre_set_beats
                    )
                self.pre_set_beats[1] = <Py_ssize_t*>realloc(
                    self.pre_set_beats[1],
                    self.num_pre_set_beats
                    )
            else:
                self.pre_set_beats[0] = <Py_ssize_t*>calloc(
                    self.num_pre_set_beats,
                    sizeof(Py_ssize_t)
                    )
                self.pre_set_beats[1] = <Py_ssize_t*>calloc(
                    self.num_pre_set_beats,
                    sizeof(Py_ssize_t)
                    )

            for n in range(self.num_pre_set_beats):
                self.pre_set_beats[0][n] = beats[n][0]
                self.pre_set_beats[1][n] = beats[n][1]

        elif prev_num_beats > 0:
            self.__dealloc__()

    def __dealloc__(self):
        if self.pre_set_beats[0] != NULL:
            free(self.pre_set_beats[0])
        if self.pre_set_beats[1] != NULL:
            free(self.pre_set_beats[1])


pd1_AC_output = OutputFuncWrapper.make_from_ptr(c_pd1_AC_output)
cdef object c_pd1_AC_output(DetectorWorkspace dws):
    """This expects a `PD1Workspace` input"""
    cdef:
        PD1Workspace ws = <PD1Workspace> dws
        frequency_info_t *f1,
        frequency_info_t *f2
        Py_ssize_t i, j, k
        Py_ssize_t m # mode index
        complex_t tmp
        Py_ssize_t Nf = ws.sim.carrier.optical_frequencies.size
        double demod_phase = (ws.phase_offset + ws.v.phase) * finesse.constants._DEG2RAD
    ws.Aij = 0

    for i in range(Nf):
        f1 = &ws.sim.carrier.optical_frequencies.frequency_info[i]
        for k in range(ws.num_unmasked_HOMs):
            m = ws.unmasked_mode_indices[k]
            tmp = ws.sim.carrier.get_out_fast(ws.dc_node_id, f1.index, m)
            ws.Aij += conj(tmp) * ws.sim.signal.get_out_fast(ws.ac_node_id, f1.audio_upper_index, m)
            # lower sideband conjugate is modelled
            ws.Aij += tmp * ws.sim.signal.get_out_fast(ws.ac_node_id, f1.audio_lower_index, m)

    # Factor of two because of signal scaling
    if ws.output_real:
        tmp = ws.Aij*cexp(-1j * demod_phase)
        return 2 * tmp.real * ws.sim.model_data.EPSILON0_C * 0.5
    else:
        return 2 * ws.Aij * ws.sim.model_data.EPSILON0_C * 0.5


pd1_DC_output = OutputFuncWrapper.make_from_ptr(c_pd1_DC_output)
cdef object c_pd1_DC_output(DetectorWorkspace dws):
    """This expects a `PD1Workspace` input"""
    cdef:
        PD1Workspace ws = <PD1Workspace> dws
        frequency_info_t *f1
        frequency_info_t *f2
        Py_ssize_t i, j, k, n
        Py_ssize_t m # mode index
        complex_t tmp
        Py_ssize_t Nf = ws.sim.carrier.optical_frequencies.size
        double demod_phase = (ws.phase_offset + ws.v.phase) * finesse.constants._DEG2RAD

    ws.Aij = 0

    if ws.num_pre_set_beats > 0:
        for n in range(ws.num_pre_set_beats):
            # Get the RHS view indices
            i = ws.pre_set_beats[0][n]
            j = ws.pre_set_beats[1][n]
            for k in range(ws.num_unmasked_HOMs):
                m = ws.unmasked_mode_indices[k]
                ws.Aij += ws.sim.carrier.out_view[i+m] * conj(ws.sim.carrier.out_view[j+m])
    else:
        # Do dumb thing and check everything if we have/can not optimise
        for i in range(ws.sim.carrier.optical_frequencies.size):
            f1 = &ws.sim.carrier.optical_frequencies.frequency_info[i]
            for j in range(ws.sim.carrier.optical_frequencies.size):
                f2 = &ws.sim.carrier.optical_frequencies.frequency_info[j]
                if float_eq(f1.f - f2.f, ws.v.f):
                    for k in range(ws.num_unmasked_HOMs):
                        m = ws.unmasked_mode_indices[k]
                        ws.Aij += (ws.sim.carrier.get_out_fast(ws.dc_node_id, f1.index, m) *
                            conj(ws.sim.carrier.get_out_fast(ws.dc_node_id, f2.index, m)))

    if demod_phase:
        tmp = ws.Aij*cexp(-1j * demod_phase)
    else:
        tmp = ws.Aij

    if ws.output_real:
        return tmp.real * ws.sim.model_data.EPSILON0_C * 0.5
    else:
        return tmp * ws.sim.model_data.EPSILON0_C * 0.5


### PD2 workspace & output funcs ###

cdef class PD2Values(BaseCValues):
    def __init__(self):
        cdef ptr_tuple_4 ptr = (&self.f1, &self.phase1, &self.f2, &self.phase2)
        cdef tuple params = ("f1", "phase1", "f2", "phase2")
        self.setup(params, sizeof(ptr), <double**>&ptr)


cdef class PD2Workspace(MaskedDetectorWorkspace):
    def __init__(self, owner, sim):
        super().__init__(owner, sim, PD2Values())
        self.v = <PD2Values>self.values

pd2_DC_output = OutputFuncWrapper.make_from_ptr(c_pd2_DC_output)
cdef object c_pd2_DC_output(DetectorWorkspace dws):
    """This expects a `PD2Workspace` input"""
    cdef:
        PD2Workspace ws = <PD2Workspace> dws
        frequency_info_t *f1
        frequency_info_t *f2
        Py_ssize_t i, j, k, n
        Py_ssize_t m # mode index
        complex_t tmp, phs
        double f_plus = ws.v.f1 + ws.v.f2
        double f_minus = ws.v.f2 - ws.v.f1
        Py_ssize_t Nf = ws.sim.carrier.optical_frequencies.size

    ws.z1 = 0
    ws.z2 = 0

    for i in range(Nf):
        f1 = &ws.sim.carrier.optical_frequencies.frequency_info[i]
        for j in range(Nf):
            f2 = &ws.sim.carrier.optical_frequencies.frequency_info[j]
            if f1.f - f2.f == f_plus:
                for k in range(ws.num_unmasked_HOMs):
                    m = ws.unmasked_mode_indices[k]
                    ws.z1 += (ws.sim.carrier.get_out_fast(ws.dc_node_id, f1.index, m) *
                        conj(ws.sim.carrier.get_out_fast(ws.dc_node_id, f2.index, m)))

            if f1.f - f2.f == f_minus:
                for k in range(ws.num_unmasked_HOMs):
                    m = ws.unmasked_mode_indices[k]
                    ws.z2 += (ws.sim.carrier.get_out_fast(ws.dc_node_id, f1.index, m) *
                        conj(ws.sim.carrier.get_out_fast(ws.dc_node_id, f2.index, m)))

    if ws.v.phase1:
        phs = cexp(-1j * ws.v.phase1 * finesse.constants._DEG2RAD)
        tmp = ws.z1 * phs + ws.z2 * conj(phs)
    else:
        tmp = ws.z1 + ws.z2

    if ws.v.phase2:
        tmp *= cexp(-1j * ws.v.phase2 * finesse.constants._DEG2RAD)

    # Factor 1/2 from demod gain
    # And a mystery factor of two to match up with v2
    if ws.output_real:
        return tmp.real * ws.sim.model_data.EPSILON0_C * 0.5 * 0.5
    else:
        return tmp * ws.sim.model_data.EPSILON0_C * 0.5 * 0.5


pd2_AC_output = OutputFuncWrapper.make_from_ptr(c_pd2_AC_output)
cdef object c_pd2_AC_output(DetectorWorkspace dws):
    """This expects a `PD2Workspace` input"""
    cdef:
        PD2Workspace ws = <PD2Workspace> dws
        frequency_info_t *f1
        frequency_info_t *f2
        Py_ssize_t i, j, k
        Py_ssize_t m # mode index
        complex_t tmp, phs, car1, car2
        double df
        double f_plus = ws.v.f1 + ws.v.f2
        double f_minus = ws.v.f2 - ws.v.f1
        Py_ssize_t Nf = ws.sim.carrier.optical_frequencies.size

    ws.z1 = 0
    ws.z2 = 0

    # Here we loop over each of the carrier beats and apply the relevant signal sideband
    # beats. This should be the equivalent code in readout elements but they fill the maxtrix
    # instead.
    for i in range(Nf):
        f1 = &ws.sim.carrier.optical_frequencies.frequency_info[i]
        for j in range(Nf):
            f2 = &ws.sim.carrier.optical_frequencies.frequency_info[j]
            df = f1.f - f2.f
            if float_eq(df, -ws.v.f1):
                for k in range(ws.num_unmasked_HOMs):
                    m = ws.unmasked_mode_indices[k]
                    car1 = ws.sim.carrier.get_out_fast(ws.dc_node_id, f1.index, m)
                    ws.z2 += car1 * ws.sim.signal.get_out_fast(ws.ac_node_id, f2.audio_lower_index, m)

                    car2 = ws.sim.carrier.get_out_fast(ws.dc_node_id, f2.index, m)
                    ws.z2 += ws.sim.signal.get_out_fast(ws.ac_node_id, f1.audio_upper_index, m) * conj(car2)

            elif float_eq(df, ws.v.f1):
                for k in range(ws.num_unmasked_HOMs):
                    m = ws.unmasked_mode_indices[k]
                    car1 = ws.sim.carrier.get_out_fast(ws.dc_node_id, f1.index, m)
                    ws.z1 += car1 * ws.sim.signal.get_out_fast(ws.ac_node_id, f2.audio_lower_index, m)

                    car2 = ws.sim.carrier.get_out_fast(ws.dc_node_id, f2.index, m)
                    ws.z1 += ws.sim.signal.get_out_fast(ws.ac_node_id, f1.audio_upper_index, m) * conj(car2)

    phs = cexp(-1j * ws.v.phase1 * finesse.constants._DEG2RAD)
    # 1/2 from cos(X)*cos(Y) = 1/2(cos(X+Y)+cos(X-Y)) expansion
    tmp = (ws.z1 * phs + ws.z2 * conj(phs)) * 0.5

    if ws.v.phase2:
        tmp *= cexp(-1j * ws.v.phase2 * finesse.constants._DEG2RAD)

    # Factor of two because of signal scaling
    if ws.output_real:
        return 2 * tmp.real * ws.sim.model_data.EPSILON0_C * 0.5
    else:
        return 2 * tmp * ws.sim.model_data.EPSILON0_C * 0.5


# ### Generic (slower) PD workspace & output funcs ###
# TODO ddb this all needs changing over to use the CarrierSignalMatrixSimulation
# class, which as none of this is used anywhere really I can't be bothered doing
# right now, as I'm not sure if any of this have any valid test cases
# cdef class PDWorkspace(DetectorWorkspace):
#     cdef:
#         public BaseSimulation DC
#         public BaseSimulation AC
#         public Py_ssize_t dc_node_id
#         public Py_ssize_t ac_node_id
#         public int[:, ::1] homs
#         const int[:, ::1] mask
#         bint mode_is_network
#         double[:, :, :, ::1] beat_factors
#         Py_ssize_t num_demod
#         Py_ssize_t num_mixes
#         double[:, ::1] fmix_view
#         double EPSILON0_C
#         complex_t[::1] zview

#     def __init__(self,
#                 owner,
#                 sims,
#                 int N_demods,
#                 double EPSILON0_C,
#                 const int[:, ::1] mask,
#                 bint mode_is_network,
#                 beat_factors=None):

#         super().__init__(owner, sims)
#         self.EPSILON0_C = EPSILON0_C
#         self.mask = mask
#         self.mode_is_network = mode_is_network
#         self.beat_factors = beat_factors
#         # Computing values and allocating arrays that will be used throughout
#         # the simulation output calculations
#         self.num_demod = N_demods
#         self.num_mixes = 1 + self.num_demod * (self.num_demod - 1) / 2
#         self.fmix_view = np.ones((self.num_mixes, self.num_demod))
#         self.zview = np.zeros(self.num_mixes, dtype=np.complex128)
#         # TODO ddb - I think the mix table only needs to be made once
#         # Might need to update if the demod frequencies change
#         _create_mix_table(self.num_demod, self.fmix_view)

# def mixer_pd_output(PDWorkspace ws,
#                     double[::1] freqs,
#                     double[::1] phases,
#                     long long[::1] max_phase_indices
# ):
#     cdef:
#         Py_ssize_t mix_index
#         double f_ref
#         complex_t dm_amp
#         complex_t power
#         # Flag for whether this pd is computing purely a transfer function
#         # TODO: Is this really needed, or just an artifact from the past?
#         int transfer = 0

#     for mix_index in range(ws.num_mixes):
#         f_ref = _get_reference_freq(ws.num_demod, mix_index, ws.fmix_view, freqs)
#         if ws.beat_factors is not None:
#             ws.zview[mix_index] = compute_amplitude_beats(ws, &transfer, f_ref)
#         else:
#             ws.zview[mix_index] = compute_amplitude(ws, &transfer, f_ref)

#     dm_amp = _set_demodulation_phase(
#         ws.num_demod, ws.zview, ws.fmix_view, ws.num_mixes, phases, max_phase_indices
#     )
#     power = _get_demodulation_signal(
#         ws.num_demod, dm_amp, phases, max_phase_indices, ws.EPSILON0_C, ws.mode_is_network, transfer
#     )

#     return power


# cdef _create_mix_table(Py_ssize_t num_demod, double[:, ::1] fmix_view):
#     cdef:
#         int num_mixes = 1

#         Py_ssize_t demod_index_outer, demod_index_inner, mix_permutation_index

#         np.ndarray[double, ndim=1] fmix_tmp = np.ones(num_demod)
#         double[::1] fmix_tmp_view = fmix_tmp

#     for demod_index_outer in range(1, num_demod):
#         for mix_permutation_index in range(demod_index_outer, 0, -1):
#             num_mixes += 1
#             fmix_tmp_view[mix_permutation_index - 1] *= -1

#             for demod_index_inner in range(num_demod):
#                 fmix_view[num_mixes - 1][demod_index_inner] = fmix_tmp_view[demod_index_inner]


# def dc_pd_output(PDWorkspace ws):
#     cdef:
#         complex_t power = COMPLEX_0
#         int transfer = 0

#     if ws.beat_factors is not None:
#         power = compute_amplitude_beats(ws, &transfer)
#     else:
#         power = compute_amplitude(ws, &transfer)

#     return power.real * ws.EPSILON0_C * 0.5


# cdef double _get_reference_freq(
#     Py_ssize_t num_demod,
#     Py_ssize_t mix_index,
#     double[:, ::1] fmix_view,
#     double[::1] freqs
# ) nogil:
#     cdef:
#         double f_ref = 0.0
#         Py_ssize_t demod_index

#     for demod_index in range(num_demod):
#         f_ref += fmix_view[mix_index][demod_index] * freqs[demod_index]

#     return f_ref



# cdef complex_t _set_demodulation_phase(
#     Py_ssize_t num_demod,
#     complex_t[::1] zview,
#     double[:, ::1] fmix_view,
#     Py_ssize_t num_mixes,
#     double[::1] phases,
#     long long[::1] max_phase_indices
# ) nogil:
#     cdef:
#         Py_ssize_t demod_index, mix_index
#         complex_t demod_amp = COMPLEX_0

#         complex_t z1, z2
#         double phi_max

#         double* phi = <double*> calloc(num_mixes, sizeof(double))

#     for demod_index in range(num_demod - 1):
#         z1 = COMPLEX_0
#         z2 = COMPLEX_0
#         # max phase
#         if in_index_array(max_phase_indices, demod_index):
#             for mix_index in range(num_mixes):
#                 if fmix_view[mix_index][demod_index] < 0.0:
#                     z1 += zview[mix_index]
#                 else:
#                     z2 += zview[mix_index]

#             phi_max = 0.5 * (carg(z2) - carg(z1))
#             phases[demod_index] = phi_max

#             for mix_index in range(num_mixes):
#                 phi[mix_index] -= phi_max * fmix_view[mix_index][demod_index]
#         # user defined phase
#         else:
#             for mix_index in range(num_mixes):
#                 phi[mix_index] -= (
#                     phases[demod_index] * fmix_view[mix_index][demod_index]
#                 )

#     for mix_index in range(num_mixes):
#         demod_amp += crotate(zview[mix_index], phi[mix_index])

#     free(phi)

#     return demod_amp


# cdef complex_t _get_demodulation_signal(
#     Py_ssize_t num_demod,
#     complex_t demod_amp,
#     double[::1] phases,
#     long long[::1] max_phase_indices,
#     double epsilon0_c,
#     bint mode_is_network,
#     int transfer
# ) nogil:
#     cdef:
#         complex_t z = demod_amp

#     if not mode_is_network:
#         if in_index_array(max_phase_indices, num_demod - 1):
#             z = cabs(demod_amp)
#         else:
#             z = (
#                 creal(demod_amp) * cos(phases[num_demod - 1])
#                 + cimag(demod_amp) * sin(phases[num_demod - 1])
#             )

#     z *= 0.5**(num_demod - 1)

#     if transfer == 2:
#         z *= 2

#     return z * epsilon0_c * 0.5

# cdef complex_t compute_amplitude(
#     PDWorkspace ws,
#     int *transfer,
#     double f_ref=NAN
# ):
#     cdef:
#         complex_t amplitude = COMPLEX_0

#         Py_ssize_t field_idx
#         Py_ssize_t N = ws.homs.shape[0]
#         int n, m
#         bint mixed = False

#     if not isnan(f_ref):
#         mixed = True

#     for field_idx in range(N):
#         n = ws.homs[field_idx][0]
#         m = ws.homs[field_idx][1]

#         if in_mask(n, m, ws.mask):
#             continue

#         amplitude += compute_mode_amplitude(
#                 ws.sim.carrier,
#                 ws.sim.signal,
#                 ws.dc_node_id,
#                 ws.ac_node_id,
#                 field_idx,
#                 field_idx,
#                 transfer,
#                 f_ref,
#                 mixed
#         )

#     return amplitude


# cdef complex_t compute_amplitude_beats(
#     PDWorkspace ws,
#     int *transfer,
#     double f_ref=NAN
# ):
#     cdef:
#         complex_t amplitude = COMPLEX_0

#         Py_ssize_t N = ws.homs.shape[0]
#         Py_ssize_t field_idx_outer, field_idx_inner
#         int n1, m1, n2, m2
#         double pd_factor
#         bint mixed = False

#     if ~isnan(f_ref):
#         mixed = True

#     for field_idx_outer in range(N):
#         n1 = ws.homs[field_idx_outer][0]
#         m1 = ws.homs[field_idx_outer][1]

#         # TODO ddb : probably worth changing ws.mask to a list of included
#         # mode indicies rather than checking all the time
#         if in_mask(n1, m1, ws.mask):
#             continue

#         for field_idx_inner in range(N):
#             n2 = ws.homs[field_idx_inner][0]
#             m2 = ws.homs[field_idx_inner][1]

#             pd_factor = ws.beat_factors [n1][m1][n2][m2]
#             if pd_factor == 0.0:
#                 continue

#             amplitude += compute_mode_amplitude(
#                     ws.sim.carrier,
#                     ws.sim.signal,
#                     ws.dc_node_id,
#                     ws.ac_node_id,
#                     field_idx_outer,
#                     field_idx_inner,
#                     transfer,
#                     f_ref,
#                     mixed
#             ) * pd_factor

#     return amplitude


# cdef complex_t compute_mode_amplitude(
#     BaseSimulation DC,
#     BaseSimulation AC,
#     Py_ssize_t dc_node_id,
#     Py_ssize_t ac_node_id,
#     Py_ssize_t field_idx_outer,
#     Py_ssize_t field_idx_inner,
#     int *transfer,
#     double f_ref,
#     bint mixed
# ) nogil:
#     cdef:
#         Py_ssize_t i, j

#         complex_t inner, outer
#         complex_t amplitude = COMPLEX_0

#         frequency_info_t f1, f2

#     for i in range(DC.num_frequencies):
#         f1 = DC.frequency_info[i]
#         outer = DC.get_out_fast(dc_node_id, f1.index, field_idx_outer)
#         for j in range(DC.num_frequencies):
#             f2 = DC.frequency_info[j]
#             if mixed:
#                 if not float_eq(f1.f - f2.f, f_ref):
#                     continue
#             else:
#                 if not float_eq(f1.f, f2.f):
#                     continue

#             transfer[0] |= 1
#             inner = DC.get_out_fast(dc_node_id, f2.index, field_idx_inner)
#             amplitude += outer * conj(inner)

#         if not mixed or AC is None:
#             continue

#         for j in range(AC.num_frequencies):
#             f2 = AC.frequency_info[j]

#             if not float_eq(f1.f - f2.f, f_ref):
#                 continue

#             transfer[0] |= 2

#             inner = AC.get_out_fast(ac_node_id, f2.index, field_idx_inner)

#             if f2.audio_order < 0:
#                 amplitude += outer * inner
#             else:
#                 amplitude += outer * conj(inner)

#     if AC is not None and mixed:
#         for i in range(AC.num_frequencies):
#             f1 = AC.frequency_info[i]
#             outer = AC.get_out_fast(ac_node_id, f1.index, field_idx_outer)
#             if f1.audio_order < 0:
#                 outer = conj(outer)
#             for j in range(DC.num_frequencies):
#                 f2 = DC.frequency_info[j]

#                 if not float_eq(f1.f - f2.f, f_ref):
#                     continue

#                 transfer[0] |= 2

#                 inner = DC.get_out_fast(dc_node_id, f2.index, field_idx_inner)

#                 amplitude += outer * conj(inner)

#     return amplitude


# cdef bint in_index_array(long long[::] indices, Py_ssize_t idx) nogil:
#     cdef:
#         Py_ssize_t i
#         Py_ssize_t N = indices.shape[0]

#     for i in range(N):
#         if indices[i] == idx:
#             return True

#     return False



#cython: boundscheck=False, wraparound=False, initializedcheck=False

"""Internal Cythonised tools for performing the calculations required by each
function in :mod:`.tracing.tools`.

Note that the functions documented here are typically only to be used as
a developer reference. Users should instead refer to :mod:`.tracing.tools`
for the Python functions which provide beam propagation tools which return
useful solution objects.
"""

cimport numpy as np
import numpy as np

from finesse.cymath cimport complex_t
from finesse.cymath.math cimport degrees
from finesse.cymath.gaussbeam cimport (
    transform_q,
    abcd_multiply,
    sym_abcd_multiply,
    gouy,
)
from finesse.tracing.ctracer cimport TraceTree

from finesse.gaussian import transform_beam_param
from finesse.utilities import refractive_index # For getting symbolic nr at a node

### Composite ABCD matrices ###


cpdef np.ndarray[double, ndim=2] compute_numeric_abcd(TraceTree last_branch, unicode direction):
    cdef:
        np.ndarray[double, ndim=2] M = np.eye(2)
        double[:, ::1] M_view = M

        bint is_x_plane = direction == "x"

        TraceTree t = last_branch.parent

    while t is not None:
        if is_x_plane:
            abcd_multiply(
                M_view, t.left_abcd_x, out=M_view,
                m2_refl_transform=t.is_left_surf_refl,
            )
        else:
            abcd_multiply(
                M_view, t.left_abcd_y, out=M_view,
            )

        t = t.parent

    return M

cpdef np.ndarray[object, ndim=2] compute_symbolic_abcd(TraceTree last_branch, unicode direction):
    cdef:
        np.ndarray[object, ndim=2] M = np.eye(2, dtype=object)
        object[:, ::1] M_view = M

        bint is_x_plane = direction == "x"

        TraceTree t = last_branch.parent

    while t is not None:
        if is_x_plane:
            sym_abcd_multiply(
                M_view, t.sym_left_abcd_x, out=M_view,
                m2_refl_transform=t.is_left_surf_refl,
            )
        else:
            sym_abcd_multiply(
                M_view, t.sym_left_abcd_y, out=M_view,
            )

        t = t.parent

    return M


### Accumulated Gouy phases ###


cpdef double compute_numeric_acc_gouy(
    TraceTree t,
    complex_t q_in,
    unicode direction,
    bint deg=True,
):
    cdef:
        # Total accumulated Gouy phase
        double agouy = 0.0
        # Gouy phase at start, end of a space
        double gouy_start = 0.0
        double gouy_end = 0.0
        # Beam parameter at current node
        complex_t q = q_in

        bint is_x_plane = direction == "x"

    if t is None:
        return agouy

    if t.node.is_input:
        if t.left is None:
            return agouy

        if is_x_plane:
            q = transform_q(t.left_abcd_x, q, t.nr, t.left.nr)
        else:
            q = transform_q(t.left_abcd_y, q, t.nr, t.left.nr)

        t = t.left

    while t is not None:
        if t.node.is_input:
            gouy_end = gouy(q)

            agouy += gouy_end - gouy_start
        else:
            gouy_start = gouy(q)

        if t.left is not None:
            if is_x_plane:
                q = transform_q(t.left_abcd_x, q, t.nr, t.left.nr)
            else:
                q = transform_q(t.left_abcd_y, q, t.nr, t.left.nr)

        t = t.left

    if deg:
        return degrees(agouy)

    return agouy

cpdef object compute_symbolic_acc_gouy(
    TraceTree t,
    object q_in,
    unicode direction,
    bint deg=True,
):
    cdef:
        # Total accumulated Gouy phase
        object agouy = 0.0
        # Gouy phase at start, end of a space
        object gouy_start = 0.0
        object gouy_end = 0.0
        # Beam parameter at current node
        object q = q_in

        bint is_x_plane = direction == "x"

    if t is None:
        return agouy

    if t.node.is_input:
        if t.left is None:
            return agouy

        if is_x_plane:
            q = transform_beam_param(t.sym_left_abcd_x.base, q, t.nr, t.left.nr)
        else:
            q = transform_beam_param(t.sym_left_abcd_y.base, q, t.nr, t.left.nr)

        t = t.left

    while t is not None:
        if t.node.is_input:
            gouy_end = q.gouy()

            agouy += gouy_end - gouy_start
        else:
            gouy_start = q.gouy()

        if t.left is not None:
            if is_x_plane:
                q = transform_beam_param(t.sym_left_abcd_x.base, q, t.nr, t.left.nr)
            else:
                q = transform_beam_param(t.sym_left_abcd_y.base, q, t.nr, t.left.nr)

        t = t.left

    if deg:
        return np.degrees(agouy)

    return agouy


### Arbitrary beam propagations ###


cpdef tuple propagate_beam_numeric(
    TraceTree t,
    q_in,
    unicode direction,
):
    cdef:
        bint is_x_plane = direction == "x"
        object q = q_in
        # Beam parameters for computing Gouy phases over spaces
        complex_t q1, q2

        # Propagated distance (geometric)
        double distance = 0.0
        # Propagated distance (optical)
        double opt_distance = 0.0

        dict node_info = {}
        dict comp_info = {}

        dict info_at_node, info_at_space, info_at_comp

    while t is not None:
        space = t.node.space
        info_at_space = comp_info.get(space, {})
        comp = t.node.component
        info_at_comp = comp_info.get(comp, {})

        if t.node.is_input:
            # Space was traversed so compute the accumulated Gouy over it
            if info_at_space:
                q1 = complex(info_at_space["q_in"])
                q2 = complex(q)

                info_at_space["acc_gouy"] = degrees(gouy(q2) - gouy(q1))

                distance += space.L.value
                opt_distance += space.L.value * t.nr

            info_at_space["q_out"] = q
        else:
            info_at_space["q_in"] = q

        info_at_comp[t.node] = q
        comp_info[comp] = info_at_comp
        comp_info[space] = info_at_space

        info_at_node = {
            "q": q,
            "z": distance,
            "z_optical": opt_distance,
            "ABCD": compute_numeric_abcd(t, direction)
        }
        node_info[t.node] = info_at_node

        if t.left is not None:
            if is_x_plane:
                q = transform_beam_param(t.left_abcd_x.base, q, t.nr, t.left.nr)
            else:
                q = transform_beam_param(t.left_abcd_y.base, q, t.nr, t.left.nr)

        t = t.left

    return node_info, comp_info

cpdef tuple propagate_beam_symbolic(
    TraceTree t,
    q_in,
    unicode direction,
):
    cdef:
        bint is_x_plane = direction == "x"
        object q = q_in
        # Beam parameters for computing Gouy phases over spaces
        object q1, q2

        # Propagated distance (geometric)
        object distance = 0.0
        # Propagated distance (optical)
        object opt_distance = 0.0

        dict node_info = {}
        dict comp_info = {}

        dict info_at_node, info_at_space, info_at_comp

    while t is not None:
        space = t.node.space
        info_at_space = comp_info.get(space, {})
        comp = t.node.component
        info_at_comp = comp_info.get(comp, {})

        if t.node.is_input:
            # Space was traversed so compute the accumulated Gouy over it
            if info_at_space:
                q1 = info_at_space["q_in"]
                q2 = q

                info_at_space["acc_gouy"] = np.degrees(q2.gouy() - q1.gouy())

                distance += space.L.ref
                opt_distance += space.L.ref * refractive_index(t.node, symbolic=True)

            info_at_space["q_out"] = q
        else:
            info_at_space["q_in"] = q

        info_at_comp[t.node] = q
        comp_info[comp] = info_at_comp
        comp_info[space] = info_at_space

        info_at_node = {
            "q": q,
            "z": distance,
            "z_optical": opt_distance,
            "ABCD": compute_symbolic_abcd(t, direction)
        }
        node_info[t.node] = info_at_node

        if t.left is not None:
            if is_x_plane:
                q = transform_beam_param(t.sym_left_abcd_x.base, q, t.nr, t.left.nr)
            else:
                q = transform_beam_param(t.sym_left_abcd_y.base, q, t.nr, t.left.nr)

        t = t.left

    return node_info, comp_info



cimport numpy as np

cpdef enum ParameterState:
    Numeric = 1,
    Unresolved = 2,
    Symbolic = 3,
    NONE =  4


cdef class Parameter:
    cdef:
        object __value
        double __cvalue
        type __datatype
        object __datatype_cast # function to cast value into __datatype
        str __units
        str __name
        str __full_name
        str __description
        bint __is_tunable
        object __weakref__
        type __component_type
        public object __component
        public bint _locked
        readonly list external_setters # If set the value of this parameter is being set by some other external element and should not be user settable
        readonly ParameterState state
        readonly bint is_geometric
        object __validator
        bint __resolving
        bint __changeable_during_simulation

    cdef __cdeepcopy__(self, Parameter new, dict memo)
    cdef __update_state(self)
    cdef bint _is_changing(self)
    cdef _get_value(self)
    cdef _set_value(self, value)
    cdef int set_double_value(self, double value) except -1


cdef class GeometricParameter(Parameter):
    cdef:
        bint is_nr

    cdef void update_abcd_matrices(self)



"""Setup file."""
from setuptools import setup, find_packages
import os
import sys
import platform
from Cython.Build import build_ext, cythonize
from Cython.Distutils import Extension

SYS_NAME = platform.system()


def get_conda_paths():
    try:
        library = sys.prefix
        # library = os.environ["CONDA_PREFIX"]
        if sys.platform == "win32":
            library = os.path.join(library, "Library")
        return (os.path.join(library, "include"), os.path.join(library, "lib"))
    except KeyError:
        return (None, None)


def make_extension(relpath, **kwargs):
    import numpy as np

    root_dir = os.path.join(os.path.abspath(os.path.dirname(__file__)), "src")
    finesse_dir = os.path.join(root_dir, "finesse")

    def construct_ext_name(rp):
        names = []
        leading, trailing = os.path.split(rp)

        while trailing != "":
            names.append(trailing)

            leading, trailing = os.path.split(leading)

        names.reverse()
        if names[-1].endswith(".pyx"):
            names[-1] = names[-1].split(".")[0]

        return ".".join(names)

    # The optional arguments consisting of various directories,
    # macros, compilation args, etc. that will be passed to
    # Extension object constructor
    ext_args = {
        "include_dirs": [],
        "define_macros": [],
        "undef_macros": [],
        "library_dirs": [],
        "libraries": [],
        "runtime_library_dirs": [],
        "extra_objects": [],
        "extra_compile_args": [],
        "extra_link_args": [],
        "export_symbols": [],
        "cython_include_dirs": [],
        "cython_directives": [],
    }
    ### Setting up some global options that need to be passed ###
    ###                   to all extensions                   ###

    include_dirs = ext_args.get("include_dirs")
    # Include the src/finesse and NumPy header file directories
    include_dirs.extend([finesse_dir, np.get_include()])

    # Now ensure suitesparse headers get included
    USR_SUITESPARSE_PATH = "/usr/include/suitesparse"
    if os.path.exists(USR_SUITESPARSE_PATH):
        include_dirs.append(USR_SUITESPARSE_PATH)

    # Grab the paths to suitesparse from conda if using this
    conda_include, conda_lib = get_conda_paths()
    if conda_include is not None:
        CONDA_SUITESPARSE_PATH = os.path.join(conda_include, "suitesparse")
        if os.path.exists(CONDA_SUITESPARSE_PATH):
            include_dirs.append(CONDA_SUITESPARSE_PATH)

        include_dirs.append(conda_include)
        ext_args.get("library_dirs").append(conda_lib)

    # define_macros = ext_args.get("define_macros")
    # define_macros.extend(
    #     [
    #         # Stops numpy version warning, cython uses an older API on purpose
    #         ("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION"),
    #     ]
    # )

    extra_compile_args = ext_args.get("extra_compile_args")
    extra_compile_args.extend(
        [
            "-O3",
            # Inlined cpdef functions in finesse.cymath extensions complain
            # about not being used (they are used outside of these extensions)
            # so we suppress these warnings for the moment
            "-Wno-unused-function",
            "-Wno-unused-variable",
            # "-DCYTHON_WITHOUT_ASSERTIONS",
        ]
    )

    ### Now adding the optional extra args needed for this specific extension ###

    for arg_opt, arg_list in ext_args.items():
        extra_args = kwargs.get(arg_opt)
        if extra_args:
            if isinstance(extra_args, str):
                extra_args = [extra_args]

            arg_list += extra_args

    ext_name = "finesse." + construct_ext_name(relpath)
    sources = [os.path.join(finesse_dir, relpath)]

    return Extension(ext_name, sources, language="c", **ext_args,)


def ext_modules():
    # Argument pattern for extensions requiring OpenMP
    if SYS_NAME == "Darwin":
        open_mp_args = {
            "extra_compile_args": ["-Xpreprocessor", "-fopenmp"],
            "extra_link_args": ["-liomp5"],
        }
    else:
        open_mp_args = {"extra_compile_args": "-fopenmp", "extra_link_args": "-fopenmp"}

    # Argument pattern for extensions using KLU
    cmatrix_args = {"libraries": "klu"}

    # The argument patterns that get passed to all extensions
    default_ext_args = {}

    COVERAGE_MODE = "--coverage" in sys.argv
    if COVERAGE_MODE:
        # If we're in coverage report mode, then add the trace
        # macros to all extensions so that proper line tracing
        # is performed
        default_ext_args["define_macros"] = [
            ("CYTHON_TRACE", "1"),
            ("CYTHON_TRACE_NOGIL", "1"),
        ]

    # NOTE (sjr) Pass any extra arguments that a specific extension needs via a
    #            dict of the arg names: values here. See ext_args in make_extension
    #            function above for the options.
    ext_args = [
        ("enums.pyx", default_ext_args),
        ("cymath/complex.pyx", default_ext_args),
        ("cymath/math.pyx", default_ext_args),
        ("cymath/gaussbeam.pyx", default_ext_args),
        ("cymath/homs.pyx", {**default_ext_args, **open_mp_args}),
        ("tree.pyx", default_ext_args),
        ("constants.pyx", default_ext_args),
        ("frequency.pyx", default_ext_args),
        ("symbols.pyx", default_ext_args),
        ("parameter.pyx", default_ext_args),
        ("cyexpr.pyx", default_ext_args),
        ("element.pyx", default_ext_args),
        ("cmatrix.pyx", {**default_ext_args, **cmatrix_args}),
        ("knm/matrix.pyx", default_ext_args),
        ("knm/bayerhelms.pyx", {**default_ext_args, **open_mp_args}),
        ("simulations/base.pyx", default_ext_args),
        ("simulations/basematrix.pyx", default_ext_args),
        ("simulations/KLU.pyx", default_ext_args),
        ("components/workspace.pyx", default_ext_args),
        ("components/mechanical.pyx", default_ext_args),
        ("components/modal/*.pyx", default_ext_args),
        ("detectors/workspace.pyx", default_ext_args),
        ("detectors/compute/amplitude.pyx", default_ext_args),
        ("detectors/compute/camera.pyx", {**default_ext_args, **open_mp_args}),
        ("detectors/compute/power.pyx", {**default_ext_args, **open_mp_args}),
        ("detectors/compute/quantum.pyx", default_ext_args),
        ("detectors/compute/gaussian.pyx", default_ext_args),
        ("tracing/ctracer.pyx", default_ext_args),
        ("tracing/cytools.pyx", default_ext_args),
        ("analysis/runners.pyx", default_ext_args),
        ("solutions/base.pyx", default_ext_args),
        ("solutions/array.pyx", default_ext_args),
        ("utilities/cyomp.pyx", {**default_ext_args, **open_mp_args}),
    ]

    exts = []
    for ext_rel_path, args in ext_args:
        exts.append(make_extension(os.path.normpath(ext_rel_path), **args))

    if (SYS_NAME == "Windows") or (SYS_NAME == "Darwin"):
        num_jobs = 0
    else:
        num_jobs = os.cpu_count()

    # See https://cython.readthedocs.io/en/latest/src/userguide/source_files_and_compilation.html#compiler-directives
    # for in-depth details on the options for compiler directives
    compiler_directives = {
        # Embeds call signature in docstring of Python visible functions
        "embedsignature": True,
        # No checks are performed on division by zero (for big perfomance boost)
        "cdivision": True,
    }

    # If coverage mode is set the ensure line tracing
    # is switched on for all extensions
    if COVERAGE_MODE:
        compiler_directives["linetrace"] = True
        sys.argv.remove("--coverage")

    # If debug mode is set then ensure profiling
    # is switched on for all extensions
    DEBUG_MODE = "--debug" in sys.argv
    if DEBUG_MODE:
        compiler_directives["profile"] = True
        sys.argv.remove("--debug")

    return cythonize(
        exts,
        # Produces HTML files showing level of CPython interaction
        # per-line of each Cython extension (.pyx) file
        annotate=True,
        language_level=3,
        nthreads=num_jobs,
        compiler_directives=compiler_directives,
        gdb_debug=True,
    )


### NOTE See https://finesse.readthedocs.io/en/latest/developer/codeguide/requirements.html

REQUIRES = [
    "numpy",
    "scipy",
    "matplotlib",
    "cython",
    "sympy",
    "networkx",
    "sly",
    "h5py",
    "click",
    "click-default-group",
    "tabulate",
    "deprecated",
]

EXTRAS = {
    "dev": [
        "sphinx",
        "sphinx_rtd_theme",
        # Can't use latest sphinxcontrib-bibtex version due to a dependency incompatibility as of
        # 2020-12-14 (https://github.com/executablebooks/jupyter-book/issues/1137). The version
        # pinning can be removed once Jupyter Book > 0.8.3 is released.
        "sphinxcontrib-bibtex < 2.0.0",
        "sphinxcontrib-katex",
        "sphinxcontrib-svg2pdfconverter",
        "sphinxcontrib-programoutput",
        "jupyter-sphinx",
        "numpydoc",
        "pytest",
        "pytest-cov",
        # Can't use latest coverage version due to Cython bug as of 2020-11-26
        # (https://github.com/cython/cython/issues/3515).
        "coverage == 4.5.4",
        "pycobertura",
        "Faker",
        "black",
        "pre-commit",
        "pylint",
        "flake8",
        "flake8-bugbear",
        "doc8",
        "hypothesis",
        "reslate",
    ],
    "graphviz": ["pygraphviz",],
}

CLASSIFIERS = [
    "Development Status :: 3 - Alpha",
    "Topic :: Scientific/Engineering :: Physics",
    "Intended Audience :: Science/Research",
    "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
    "Operating System :: Microsoft :: Windows",
    "Operating System :: POSIX",
    "Operating System :: Unix",
    "Operating System :: MacOS",
]


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


def parallelise_build():
    import warnings
    from concurrent.futures import ProcessPoolExecutor
    from concurrent.futures.process import BrokenProcessPool

    def build_extensions(self):
        """Function to monkey-patch Cython.Build.build_ext.build_extensions to build
        extensions in parallel."""
        self.check_extensions_list(self.extensions)
        num_jobs = os.cpu_count()
        try:
            with ProcessPoolExecutor(max_workers=num_jobs) as pool:
                # results = pool.map(self.build_extension, self.extensions)
                pool.map(self.build_extension, self.extensions)
        # TODO [adf] check which exception could be used to start sequential build
        except BrokenProcessPool:
            for ex in self.extensions:
                self.build_extension(ex)
        # TODO [adf] this does not raise an error when the build fails, so
        # it still looks like everything works with pip install
        # else:
        #    try:
        #        print(list(results))
        #    except (Exception, BrokenProcessPool):
        #        raise RuntimeError("Could not compile extension, maybe missing a dependency")

    warnings.warn(
        "Cython is monkey-patched to parallelise the build. If you encounter any issues, this "
        "behaviour can be disabled in setup.py."
    )
    build_ext.build_extensions = build_extensions


# Parallel build behaves strangely on Windows and Mac OSX so enable it only for Linux.
# See #272.
if __name__ == "__main__" and SYS_NAME == "Linux":
    parallelise_build()


setup(
    # package descriptions
    name="finesse",
    description=("A simulation tool for modelling laser interferometers"),
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    url="https://www.gwoptics.org/finesse",
    author_email="finesse-support@nikhef.nl",
    project_urls={
        "Bug Tracker": "https://git.ligo.org/finesse/finesse3/issues",
        "Documentation": "https://finesse.docs.ligo.org/finesse3",
        "Source Code": "https://git.ligo.org/finesse/finesse3",
    },
    # packages and extensions
    packages=find_packages("src"),
    package_dir={"": "src"},
    ext_modules=ext_modules(),
    cmdclass={"build_ext": build_ext},
    zip_safe=False,
    package_data={
        "finesse": [
            "plotting/style/*.mplstyle",
            "finesse.ini",  # Base config file.
            "usr.ini.dist",  # Barebone user config file.
        ]
    },
    # requirements
    python_requires=">=3.8",
    install_requires=REQUIRES,
    setup_requires=["setuptools_scm", "numpy"],
    extras_require=EXTRAS,
    # other
    license="GPL",
    classifiers=CLASSIFIERS,
    # CLI
    entry_points={
        "console_scripts": ["kat3 = finesse.__main__:cli"],
        "pygments.lexers": [
            "KatScript = finesse.script.highlighter:KatScriptPygmentsLexer",
            "KatScriptInPython = finesse.script.highlighter:KatScriptSubstringPygmentsLexer",
        ],
    },
)



cimport cython
import numpy as np
cimport numpy as np
from enum import Enum
from copy import deepcopy
import logging
import numbers
import re
import weakref
from collections import namedtuple
from finesse.element import ModelElement
from finesse.exceptions import (
    ComponentNotConnected, ParameterLocked, ModelParameterSelfReferenceError
)
from finesse.symbols import FUNCTIONS, Symbol, Resolving


LOGGER = logging.getLogger(__name__)
c_reference = re.compile(r"([a-zA-Z][a-zA-Z0-9_]*).([a-zA-Z][a-zA-Z0-9_]*)")


# Allowed datatypes for parameters
allowed_datatypes = (int, float, bool, Enum, np.int64, np.float64, np.bool_)


# struct for storing float_parameter information on each element
ParameterInfo = namedtuple(
    "ParameterInfo", (
        "name", "description", "dtype", "dtype_cast", "units", "is_geometric", "changeable_during_simulation")
)


class ParameterRef(Symbol):
    """
    A symbolic instance of a parameter in the model. A parameter
    is owned by some model element, which can be used to connect
    its value to other aspects of the simulation.
    """

    def __init__(self, param):
        self.__param = weakref.ref(param)
        self.__name = f"{self.owner.name}.{self.parameter.name}"
        self.__cyexpr_name = self.__name.replace(".", "_").lower().encode("UTF-8")

    @property
    def parameter(self):
        return self.__param()

    @property
    def name(self):
        return self.__name

    @property
    def cyexpr_name(self):
        """Name of the parameter reference in ``cyexpr`` compatibility
        format.

        This is equivalent to :attr:`.ParameterRef.name` but with ``"."``
        replaced with ``"_"``, converted to lower case and encoded
        in UTF-8 format as a ``bytes`` object.

        The above format makes this compatible with passing to the underlying
        math evaluator engine (``tinyexpr``) used via the :mod:`.cyexpr` sub-module.

        .. note::

            This should, typically, never need to be used outside of internal usage. It
            exists primarily to act as the owner for the parameter name strings (avoiding
            dangling pointers in the expression code).

        :getter: Returns the ``cyexpr`` compatible name format of the pref name (read-only).
        """
        return self.__cyexpr_name

    @property
    def owner(self):
        return self.parameter.component

    def __eq__(self, obj):
        if isinstance(obj, ParameterRef):
            if obj.parameter is self.parameter:
                return True
        elif isinstance(obj, Parameter):
            if obj is self.parameter:
                return True

    def __hash__(self):
        return hash(self.parameter)

    def __deepcopy__(self, memo):
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        result.__dict__.update(deepcopy(self.__dict__, memo))
        owner_id = id(self.parameter.owner)
        new_model = memo.get(id(self.parameter.owner._model))

        # We need to update this reference later on
        # This will be called when the port property
        # is accessed. When this happens we'll peak back
        # at the memo once it has been filled and get
        # the new port reference. After this the refcount
        # for this function should goto zero and be garbage
        # collected
        def update_later():
            result.__param = weakref.ref(
                getattr(memo[owner_id], self.parameter.name)
            )

        result.__param = update_later  # just in case something calls this weakref in the meantime
        new_model.after_deepcopy.append(update_later)

        return result

    def eval(self, keep_changing_symbols=False, subs=None, **kwargs):
        p = self.parameter

        if keep_changing_symbols and p.is_changing:
            return self
        if subs is not None and self in subs:
            return subs[self]
        if hasattr(p.value, "eval"):
            return p.value.eval()
        return p.value


cdef class Parameter:
    def __init__(self, parameter_info, component):
        from finesse.element import ModelElement
        self.__value = None
        self.__cvalue = 0.0

        self.__units = parameter_info.units
        self.__name = parameter_info.name
        self.__datatype = parameter_info.dtype
        self.__datatype_cast = parameter_info.dtype_cast
        self.__full_name = f"{component.name}.{parameter_info.name}"
        self.__description = parameter_info.description
        self.__component = weakref.ref(component)
        self.__component_type = type(component)
        self.__changeable_during_simulation = parameter_info.changeable_during_simulation
        self._locked = False
        self.external_setters = []
        self.__is_tunable = False
        self.__update_state()
        self.is_geometric = False
        self.__validator = ModelElement._validators[self.__component_type][self.name]
        self.__resolving = False

    cdef __cdeepcopy__(self, Parameter new, dict memo):
        new.__units = self.__units
        new.__name = self.__name
        new.__datatype = self.__datatype
        new.__datatype_cast = self.__datatype_cast
        new.__full_name = self.__full_name
        new.__description = self.__description
        new._locked = self._locked
        new.external_setters = deepcopy(self.external_setters, memo)
        new.__is_tunable = self.__is_tunable
        new.state = self.state
        new.__value = deepcopy(self.__value, memo)
        new.__cvalue = self.__cvalue
        new.__component_type = self.__component_type
        new.__validator = self.__validator
        new.__changeable_during_simulation = self.__changeable_during_simulation

    def __deepcopy__(self, memo):
        new = Parameter.__new__(Parameter)
        memo[id(self)] = new
        self.__cdeepcopy__(new, memo)

        # Manually update the weakrefs to be correct
        id_component = id(self.component)

        if id_component not in memo:
            # We need to update this reference later on
            # This will be called when the port property
            # is accessed. When this happens we'll peak back
            # at the memo once it has been filled and get
            # the new port reference. After this the refcount
            # for this function should goto zero and be garbage
            # collected
            def update_later():
                new.__component = weakref.ref(memo[id(self.component)])

            new.__component = update_later  # just in case something calls
            # this weakref in the meantime
            memo[id(self.__component()._model)].after_deepcopy.append(update_later)
        else:
            new.__component = weakref.ref(memo[id(self.component)])

        return new

    cdef __update_state(self):
        if issubclass(type(self.__value), allowed_datatypes):
            self.state = ParameterState.Numeric
        elif self.__value is None:
            self.state = ParameterState.NONE
        elif isinstance(self.__value, Symbol):
            self.state = ParameterState.Symbolic
        elif callable(self.__value):
            self.state = ParameterState.Unresolved
        else:
            raise Exception(f"Unexpected parameter value '{self.__value}' ({type(self.__value))})")

    @property
    def units(self):
        if self.__units is None:
            return ""
        return self.__units

    @property
    def name(self):
        return self.__name

    @property
    def datatype(self):
        """The underlying C datatype of this parameter"""
        return self.__datatype

    def datatype_cast(self, value, ignore_failure=False):
        """Casts a value into the datatype of this parameter.
        If ignore_failure is True then if this value cannot
        be cast it is just returned.
        """
        try:
            return self.__datatype_cast(value)
        except (ValueError, TypeError) as ex:
            if ignore_failure:
                return value
            else:
                raise ex

    @property
    def component(self):
        """The component/element this parameter is associated with"""
        return self.__component()

    @property
    def owner(self):
        """The component/element this parameter is associated with"""
        return self.__component()

    @property
    def locked(self):
        """If locked, this parameters value cannot be changed"""
        return self._locked

    @property
    def changeable_during_simulation(self):
        """True if this parameter cannot be changed during a simulation"""
        return self.__changeable_during_simulation

    @property
    def _model(self):
        """
        Returns the model that this component is connected to. Raises an exception if
        it is not connected to anything to stop code accidently using unconnected
        components by accident.

        Returns
        -------
        :class:`.model.Model`

        Raises
        ------
        :class:`.exceptions.ComponentNotConnected`
        """
        if self.__component() is None:
            raise ComponentNotConnected("Lost weak reference")
        else:
            return self.__component()._model

    @property
    def full_name(self):
        return self.__full_name

    @property
    def description(self):
        return self.__description

    @property
    def ref(self):
        """Returns a reference to this parameter's value to be used in symbolic expressions."""
        return ParameterRef(self)

    cdef bint _is_changing(self):
        if self.state == ParameterState.Symbolic:
            return self.value.is_changing

        return self.__is_tunable

    @property
    def is_changing(self):
        """True if this parameter will be changing during a simulation"""
        return self._is_changing()

    @property
    def is_tunable(self):
        """True if this parameter will be directly changed during a simulation"""
        return self.__is_tunable

    @property
    def is_symbolic(self):
        """True if this parameter's value is symbolic"""
        return self.state == ParameterState.Symbolic

    @is_tunable.setter
    def is_tunable(self, bint value):
        if not self.changeable_during_simulation:
            raise Exception(f"The property {self.full_name} cannot be changed during a simulation")
        if self._locked:
            raise ParameterLocked()
        if len(self.external_setters) > 0:
            raise Exception(f"{repr(self)} is being externally controlled by {self.external_setters}, so it cannot be tuned explicitly.")
        if self.state == ParameterState.Symbolic:
            raise Exception(
                f"Tunable state of {repr(self)} cannot be set to True when its value is symbolic."
            )

        self.__is_tunable = value

    def resolve(self):
        """When this parameters value has some dependency whose value has not yet been set,
        like during parsing, its value will be a callable object. This method will call this
        function to return the value.
        """
        self.__resolving = True

        if callable(self.value):
            self.value = self.value(self.component._model)

        self.__resolving = False
        self.__update_state()

    def lambdify(self, *args):
        """Returns a lambda function that returns the value of this parameter.

        Parameters in a symbolic function can be kept as variables by passing
        the Parameter object as optional arguments. The returned lambda function
        will then have len(args) arguments - effectively subsituting values at
        call time.
        """
        if self.state == ParameterState.Symbolic:
            refs = {**self.component._model.elements, **FUNCTIONS}
            sym_str = str(self.value)
            ARGS = []
            for i, arg in enumerate(args):
                ARGS.append(arg.full_name.replace('.','__'))
                sym_str = c_reference.sub(ARGS[-1], sym_str)

            return eval(f"lambda {','.join(ARGS)}: {sym_str}", refs)
        else:
            # if len(args) > 0:
            #     raise Exception("Can't specify symbolic arguments to lambdify if this parameter isn't symbolic.")
            return lambda *x: self.value

    def eval(self, bint keep_changing_symbols=False):
        """
        Evaluates the value of this parameter. If the parameter
        is dependant on some symbolic statement this will evaluate
        that. If it is not the value itself is returned. This method
        should be used when filling in matrices for computing solutions
        of a model.
        """
        if self.state == ParameterState.Unresolved:
            self.resolve()

        if self.state == ParameterState.Symbolic:
            y = self.value.eval(keep_changing_symbols=keep_changing_symbols)
            if self.__validator is None:
                return y
            else:
                return self.__validator(self.owner, y)

        else:
            return self.value

    cdef _get_value(self):
        return self.__value

    @property
    def value(self):
        return self._get_value()

    def set_external_setter(self, element, symbol):
        """Sets an element as an external controller of the value of this parameter.
        It is used in cases where an element is physically imposing its value onto this
        one, such as degrees of freedom.

        Parameters
        ----------
        element : ModelElement
            Element that will be controlling this parameter

        symbol : symbolic expression
            The expression that this element is imposing upon the parameter.
        """
        if element in self.external_setters:
            raise Exception(f"{repr(element)} is already an external setter of {repr(self)}")
        self.external_setters.append(element)
        self._set_value(self.value + symbol)

    def remove_external_setter(self, element, symbol):
        """Stops an element from being an external setter.

        Parameters
        ----------
        element : ModelElement
            Element that is controlling this parameter
        symbol : symbolic expression
            The symbolic expression that was added to this parameters value, this will
            be subtracted.
        """
        if element not in self.external_setters:
            raise Exception(f"{repr(element)} is not an external setter of {repr(self)}")
        self.external_setters.remove(element)
        # Use sympy to simplify the expression to remove the symbol
        self._set_value((self.value - symbol).sympy_simplify())

    def _reset_cvalue(self):
        """Resets the C-level value. To be used in clean-up stage of an action."""
        self.__cvalue =  self.__datatype_cast(self.eval()) # cast to correct datatype

    cdef _set_value(self, value):
        from finesse.element import ModelElement
        setting_fn = ModelElement._setters[type(self.owner)].get(self.name)
        cdef bint is_symbol = isinstance(value, Symbol)
        cdef bint is_callable = isinstance(value, Resolving)

        if setting_fn is not None and not self.__resolving and not is_symbol:
            handle = setting_fn(self.owner)
            if not handle.has_been_called:
                oname = self.owner.name
                LOGGER.warning(
                    "Direct setting of %s.%s is not recommended. You should "
                    "use %s.%s to set this parameter.",
                    oname,
                    self.name,
                    oname,
                    handle.__name__,
                )

        if self.locked:
            if self.state == ParameterState.Symbolic:
                raise ParameterLocked(f"Parameter {repr(self)} is symbolic so value cannot be directly set")
            else:
                raise ParameterLocked(f"Parameter {repr(self)} is locked")

        if is_symbol:
            for p in value.parameters():
                if p == self.ref:
                    raise ModelParameterSelfReferenceError(value, self)

        if self.__is_tunable and self.state == ParameterState.Symbolic:
            LOGGER.warn(
                f"{repr(self)}: Setting `is_tunable` to False as new "
                "value is symbolic, is_tunable was previously True."
            )
            self.__is_tunable = False

        if is_symbol or value is None or is_callable:
            # Don't cast here as need to keep original object
            self.__value = value
        elif self.__validator is None:
            self.__value = self.__datatype_cast(value)
        else:
            self.__value = self.__validator(self.owner, self.__datatype_cast(value))

        # Correspondingly set the low-level __cvalue as long as value
        # is valid and not in an unresolved state
        if value is not None and not is_callable:
            try:
                self.__cvalue =  self.__datatype_cast(self.__value)
            # Return of self.eval is allowed to be None, unfortunately, so
            # take care of that here (any other type error would invariably
            # show up earlier than this anyway)
            except TypeError:
                pass

        self.__update_state()

    @value.setter
    def value(self, value):
        if len(self.external_setters) > 0:
            raise Exception(f"Model parameter {repr(self)} is being externally controlled by {self.external_setters}")

        self._set_value(value)

    cdef int set_double_value(self, double value) except -1:
        """
        UNSAFE!
        Only use when you know the parameter should be changing.
        """
        # TODO (sjr) Not sure setting of self.__value here is reqd
        #            anymore as just setting __cvalue should suffice
        #            here now (as this method should only be called
        #            during simulations anyway)
        if self.__validator is None:
            self.__value = value
        else:
            self.__value = self.__validator(self.owner, value)

        self.__cvalue = value

        return 0

    def _set(self, value):
        self.value = value

    def __str__(self):
        units = self.units
        if units:
            units = " " + units
        return f"{self.value}{units}"

    def __repr__(self):
        if self.__component() is None:
            return f"<LOST WEAKREF.{self.name}={str(self.value)}>"
        else:
            return f"<{self.__component().name}.{self.name}={str(self.value)} @ {hex(id(self))}>"

    def __pow__(self, power, order):
        return self.value ** power

    def __mul__(A, B):
        if isinstance(A, Parameter):
            return A.value * B
        else:
            return A * B.value

    def __add__(A, B):
        if isinstance(A, Parameter):
            return A.value + B
        else:
            return A + B.value

    def __sub__(A, B):
        if isinstance(A, Parameter):
            return A.value - B
        else:
            return A - B.value

    def __pos__(self):
        return self

    def __neg__(self):
        return -1 * self

    def __floordiv__(A, B):
        if isinstance(A, Parameter):
            return A.value // B
        else:
            return A // B.value

    def __truediv__(A, B):
        if isinstance(A, Parameter):
            return A.value / B
        else:
            return A / B.value

    def __complex__(self):
        return complex(self.value)

    def __float__(self):
        return float(self.value)

    def __int__(self):
        return int(self.value)

    def __hash__(self):
        return id(self)

    def __eq__(A, B):
        if hasattr(A, "value"):
            if hasattr(B, "value"):
                return A.value == B.value
            return A.value == B
        return A == B.value

    def __ge__(A, B):
        if hasattr(A, "value"):
            if hasattr(B, "value"):
                return A.value >= B.value
            return A.value >= B
        return A >= B.value

    def __gt__(A, B):
        if hasattr(A, "value"):
            if hasattr(B, "value"):
                return A.value > B.value
            return A.value > B
        return A > B.value

    def __le__(A, B):
        if hasattr(A, "value"):
            if hasattr(B, "value"):
                return A.value <= B.value
            return A.value <= B
        return A <= B.value

    def __lt__(A, B):
        if hasattr(A, "value"):
            if hasattr(B, "value"):
                return A.value < B.value
            return A.value < B
        return A < B.value


cdef class GeometricParameter(Parameter):
    """Specialised parameter class for variables which are dependencies
    of ABCD matrices. These include surface radii of curvature, lens focal
    lengths, beamsplitter angles of incidence, space lengths and space
    refractive indices.

    When setting the value of a GeometricParameter *outside of a simulation*, the dependent ABCD
    matrices get updated via the `update_abcd_matrices` C method. Inside a simulation,
    the ABCD matrix elements are updated in a much more efficient way via the connector
    workspaces.
    """
    def __init__(self, parameter_info, component):
        super().__init__(parameter_info, component)
        self.is_geometric = True
        self.is_nr = self.name == "nr"

    def __deepcopy__(self, memo):
        new = GeometricParameter.__new__(GeometricParameter)
        memo[id(self)] = new
        self.__cdeepcopy__(new, memo)

        # Manually update the weakrefs to be correct
        id_component = id(self.component)

        if id_component not in memo:
            # We need to update this reference later on
            # This will be called when the port property
            # is accessed. When this happens we'll peak back
            # at the memo once it has been filled and get
            # the new port reference. After this the refcount
            # for this function should goto zero and be garbage
            # collected
            def update_later():
                new.__component = weakref.ref(memo[id(self.component)])

            new.__component = update_later  # just in case something calls
            # this weakref in the meantime
            memo[id(self.__component()._model)].after_deepcopy.append(update_later)
        else:
            new.__component = weakref.ref(memo[id(self.component)])

        return new

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.initializedcheck(False)
    cdef void update_abcd_matrices(self):
        """Method for updating dependent ABCD matrices of the parameter *outside*
        of a simulation."""
        cdef:
            dict abcd_matrices = self.component._abcd_matrices
            list matrix_handles = list(abcd_matrices.values())

            object full_symbolic # Full symbolic ABCD matrix
            double[:, ::1] M_num # Numeric ABCD matrix

            object[:, ::1] M_sym # Actual symbolic matrix to use

        if self.is_nr: # Parameter is a refractive index
            from finesse.components import Surface
            if self.component.portA is not None:
                p1_comp = self.component.portA.component
            else:
                p1_comp = None
            if self.component.portB is not None:
                p2_comp = self.component.portB.component
            else:
                p2_comp = None
            # and if either / both ports are associated with Surfaces
            # then append their ABCD matrices so that they get updated
            # too (as some of them depend on nr of attached space)
            if isinstance(p1_comp, Surface):
                matrix_handles.extend(list(p1_comp._abcd_matrices.values()))
            if isinstance(p2_comp, Surface):
                matrix_handles.extend(list(p2_comp._abcd_matrices.values()))

        cdef Py_ssize_t i, j, k
        cdef Py_ssize_t N = len(matrix_handles)
        for i in range(N):
            full_symbolic = matrix_handles[i][0]
            # Total reflection means no symbolic matrix is formed
            # so ignore it and move on
            if full_symbolic is None:
                continue

            M_sym = full_symbolic
            M_num = matrix_handles[i][1]

            for j in range(2):
                for k in range(2):
                    M_num[j][k] = float(M_sym[j][k])

    # NOTE (sjr) ABCD matrices are updated using cyexpr's in the relevant
    #            workspace during a simulation, so we don't override
    #            set_double_value here. The update_abcd_matrices method
    #            is only called when using the value setter, which will be
    #            used outside of a simulation context.

    @property
    def value(self):
        return Parameter._get_value(self)

    @value.setter
    def value(self, value):
        Parameter._set_value(self, value)
        self.update_abcd_matrices()


class parameterproperty(property):
    """
    Descriptor class for declaring a simulation parameter. A simulation parameter
    is one that can be changed during a simulation and affect the resulting outputs.
    The idea is that output dependant variables should be marked as having been changed
    or will be changed during a simulation run. This allows us to then optimise parts of
    the model, as we can determine what will or will not be changing. This descriptor is
    paired with the :class:Parameter.

    Parameters can then be superficially locked once a model has been built so accidentally
    changing some parameter that isn't expected to change can flag a warning.
    """

    def __init__(self, fget, fset, flocked, doc=None):
        super().__init__(fget, fset, doc=doc)

        self.fget.__doc__ = ":getter: " + doc
        self.fset.__doc__ = ":setter: " + doc
        self.flocked = flocked
        self.flocked.__doc__ = "Parameter locking function for " + doc

    def __get__(self, obj, objtype=None):
        if obj is None:
            return self
        else:
            return self.fget(obj)

    def __set__(self, obj, value):
        if self.flocked(obj):
            raise ParameterLocked()
        else:
            self.fset(obj, value)

    def __delete__(self, obj):
        raise AttributeError("can't delete parameter")

    def getter(self, fget):
        return type(self)(fget, self.fset, self.fdel, self.__doc__)

    def setter(self, fset):
        return type(self)(self.fget, fset, self.fdel, self.__doc__)


def float_parameter(
    name, description, units=None, validate=None, setter=None,
    is_default=False, is_geometric=False, changeable_during_simulation=True
):
    """A parameter of an element whose value is decribed by a 64-bit floating point number"""
    return model_parameter(
        name, description, float, float,
        units=units, validate=validate, setter=setter,
        is_default=is_default, is_geometric=is_geometric,
        changeable_during_simulation=changeable_during_simulation
    )


def int_parameter(
    name, description, units=None, validate=None, setter=None,
    is_default=False, is_geometric=False, changeable_during_simulation=True
):
    """A parameter of an element whose value is decribed by a 64-bit integer number"""
    return model_parameter(
        name, description, int, int,
        units=units, validate=validate, setter=setter,
        is_default=is_default, is_geometric=is_geometric,
        changeable_during_simulation=changeable_during_simulation
    )


def enum_parameter(
    name, description, enum, units=None, validate=None, setter=None,
    is_default=False, is_geometric=False, changeable_during_simulation=True
):
    """A parameter of an element whose value is decribed by a Enum definition.
    Enum must only use integer values to describe its members. Unlike a general
    python Enum which can use strings.
    """
    if not issubclass(enum, Enum):
        raise Exception("enum argument should be an enum.Enum type object")

    for _ in enum.__members__.values():
        if type(_.value) is not int:
            raise Exception(f"Value for item {repr(_)} in {enum} must be an integer")

    # enums are irritating as they don't act like normal types. You need
    # to getitem to go from value to key. They also don't cast themselves
    # back to themselves as float(1.0) would.
    def enum_cast(value):
        if type(value) is enum:
            return value
        else:
            return enum[value]

    return model_parameter(
        name, description, int, enum_cast,
        units=units, validate=validate, setter=setter,
        is_default=is_default, is_geometric=is_geometric,
        changeable_during_simulation=changeable_during_simulation
    )


def bool_parameter(
    name, description, units=None, validate=None, setter=None,
    is_default=False, is_geometric=False, changeable_during_simulation=True
):
    """A parameter of an element whose value is decribed by a True or False value"""
    return model_parameter(
        name, description, bool, bool,
        units=units, validate=validate, setter=setter,
        is_default=is_default, is_geometric=is_geometric,
        changeable_during_simulation=changeable_during_simulation
    )


cdef object model_parameter(
    str name, str description, type _type, object _type_cast, str units, str validate, str setter,
    bint is_default, bint is_geometric, bint changeable_during_simulation
):
    """Decorator to register a model parameter with a double datatype field in the class.
    This shouldn't be called by users. Use the [type]_parameter functions above instead.
    """
    if not issubclass(_type, allowed_datatypes):
        raise Exception(f"Data type {_type} not allowed ({allowed_datatypes})")

    if validate is None:
        vld = None
    else:
        vld = lambda x, v: getattr(x, validate)(v)

    # setter is the name of a method which will be the only method that can
    # be used to set the parameter (direct setting will be forbidden)
    if setter is None:
        st = None
    else:
        st = lambda x: getattr(x, setter)

    def func(cls):
        p = parameterproperty(
            lambda x: getattr(x, f"__param_{name}"),
            lambda x, v: getattr(x, f"__param_{name}")._set(v),
            lambda x: getattr(x, f"__param_{name}").locked,
            doc=description
        )
        cls._param_dict[cls].append(
            ParameterInfo(
                name, description, _type, _type_cast, units, is_geometric, changeable_during_simulation
            )
        )
        cls._validators[cls][name] = vld
        cls._setters[cls][name] = st

        if is_default:
            if cls in cls._default_parameter_name:
                raise ValueError(
                    f"is_default cannot be set for more than one model parameter on {cls!r}"
                )

            cls._default_parameter_name[cls] = name

        setattr(cls, name, p)
        return cls

    return func


def info_parameter(name, description, units=None):
    """Decorator to register an info parameter field in the class.

    Info parameters are purely informative properties, and cannot be directly scanned using an axis.
    """
    def func(cls):
        cls._info_param_dict[cls][name] = (description, units)
        return cls

    return func



from finesse.cmatrix cimport SubCCSView, SubCCSView2DArray
from finesse.cymath.complex cimport complex_t
from finesse.components.workspace cimport ConnectorWorkspace, FillFuncWrapper
from finesse.element cimport BaseCValues

import numpy as np
cimport numpy as np
from cpython.ref cimport PyObject

cdef struct squeezer_noise_sources:
    # 2D array of SubCCSViews
    PyObject*** P1o

cdef class SqueezerConnections:
    cdef public:
        int P1o_idx
    cdef readonly:
        SubCCSView2DArray P1o

cdef class SqueezerNoiseSources:
    cdef readonly:
        SubCCSView2DArray P1o
    cdef:
        squeezer_noise_sources ptrs

cdef class SqueezerValues(BaseCValues):
    cdef public:
        double db
        double angle
        double f


cdef class SqueezerWorkspace(ConnectorWorkspace):
    cdef public:
        Py_ssize_t fsrc_car_idx
        Py_ssize_t node_id
        SqueezerValues v
        SqueezerNoiseSources ns
        complex_t[:, ::1] qn_coeffs
        complex_t[::1] qn_coeffs_diag


cdef object c_squeezer_fill_rhs(ConnectorWorkspace cws)



# cython profile=True

import contextlib
import logging
import math
import weakref
import re
import numpy as np
cimport numpy as np

from libc.stdlib cimport free, malloc, calloc

cimport cython

from finesse.cmatrix cimport CCSMatrix, SubCCSView
from finesse.detectors.general import NoiseDetector
from finesse.components.general import NoiseType
from finesse.components.readout import Readout
from finesse.components.node import NodeType
from finesse.simulations.base cimport NodeInfoEntry
from finesse.frequency cimport Frequency, frequency_info_t, FrequencyContainer
from finesse.components.workspace cimport ConnectionSetting
from finesse.cymath cimport complex_t
from finesse.cymath.math cimport float_eq
from finesse.cymath.complex cimport conj
from finesse.cymath.gaussbeam cimport beam_param, transform_q, inv_transform_q
from finesse.detectors.workspace import DetectorWorkspace
from finesse.element cimport ElementWorkspace
from finesse.components.workspace cimport ConnectorWorkspace, FillFuncWrapper, ConnectorMatrixSimulationInfo
from finesse.components.modal.workspace cimport KnmConnectorWorkspace
from finesse.components.workspace cimport ConnectorCallbacks, fill_list_t
from finesse.components.modal.cavity cimport CavityWorkspace

from finesse.utilities.cyomp cimport determine_nthreads_even

LOGGER = logging.getLogger(__name__)


cdef class MatrixSystemSolver:
    """A linear set of systems can be represented as a matrix, each equation
    in this system is a particular state which we want to compute. The system
    is solved by applying some inputs into various states, or the right hand
    side (RHS) vector, and solving the system.

    The underlying matrix can be either a sparse or dense matrix. This class
    should not assume either, but merely call upon a standard matrix interface.
    Therefore the algorithm used for solving can varying significantly.
    The overall matrix is sectioned into submatricies which connect various
    states together.

    Nodes represent a physical location in the model in which some state of
    the system must be computed. Some nodes can have multiple states, such
    as multiple optical modes.
    """

    def __init__(self,
                matrix_type,
                str name,
                list nodes,
                FrequencyContainer optical_frequencies,
                dict mechanical_frequencies,
                dict electrical_frequencies,
                int num_optical_homs,
                bint is_signal_matrix,
                bint forced_refill,
    ):
        if is_signal_matrix:
            if (mechanical_frequencies is None or electrical_frequencies is None):
                raise Exception("Mechanical and electrical frequency containers not provided")
        else:
            if mechanical_frequencies is not None or electrical_frequencies is not None:
                raise Exception("Mechanical and electrical frequency containers incorrectly provided for carrier simulation")
        assert(num_optical_homs >= 1)

        self.is_signal_matrix = is_signal_matrix
        self.manual_rhs = False
        self.workspaces = MatrixSystemWorkspaces()
        self.forced_refill = forced_refill
        self._M = matrix_type(name)
        self.out_view = None
        self.nhoms = num_optical_homs
        self.optical_frequencies = optical_frequencies
        self.mechanical_frequencies = mechanical_frequencies
        self.electrical_frequencies = electrical_frequencies

        if is_signal_matrix:
            # Get the unique FrequencyContainer objects for filling the info of.
            # This could be more optimal and dive into the references and check if
            # frequency containers are actually unique, but this mostly just stops
            # there being 100s of fsig containers from each node
            # TODO - optimise this away from a tuple
            self.unique_elec_mech_fcnts = tuple(
                set(mechanical_frequencies.values()) |
                set(electrical_frequencies.values())
            )
            self.noise_sources = dict()
            self._noise_matrices = dict()

        self.any_frequencies_changing = False
        # Global flag for if any frequency is changing
        for _f in self.optical_frequencies.frequencies:
            if _f.symbol.is_changing:
                self.any_frequencies_changing = True
                break

        self.setup_nodes(nodes)

    def __dealloc__(self):
        if self._c_node_info != NULL:
            free(self._c_node_info)

    cdef get_node_matrix_params(self, node, Py_ssize_t *Ns, Py_ssize_t *Nf, frequency_info_t** fptr):
        """For a given node in the simulation this should set the provided pointers
        regarding the number of states and submatricies that are required in the matrix:
            - Ns : Number of unique states at the node per frequency
            - Nf : Number of frequencies at the node
            - fptr : Pointer to frequency_info_t for details on the number of frequencies
        """
        assert(Ns)
        assert(Nf)
        assert(fptr)

        cdef FrequencyContainer ficnt

        if node.type is NodeType.OPTICAL:
            Ns[0] = self.nhoms
            Nf[0] = self.optical_frequencies.size
            fptr[0] = &self.optical_frequencies.frequency_info[0]
        elif node.type is NodeType.MECHANICAL:
            # Higher order mechanical modes at a particular frequency. This should probably
            # be kept as 1 mode per frequency, additional mechanical degrees of freedom should
            # be defined as a separate node in a port.
            ficnt = <FrequencyContainer>(self.mechanical_frequencies[node])
            Ns[0] = 1
            Nf[0] = ficnt.size
            fptr[0] = &(ficnt.frequency_info[0])
        elif node.type is NodeType.ELECTRICAL:
            ficnt = <FrequencyContainer>(self.electrical_frequencies[node])
            Ns[0] = 1 # no higher order modes of electronics as far as I'm aware...
            Nf[0] = ficnt.size
            fptr[0] = &(ficnt.frequency_info[0])
        else:
            raise Exception("Node type not handled")

    cdef tuple get_node_frequencies(self, node):
        if node.type == NodeType.OPTICAL:
            return self.optical_frequencies.frequencies
        elif node.type == NodeType.MECHANICAL:
            return self.mechanical_frequencies[node].frequencies
        elif node.type == NodeType.ELECTRICAL:
            return self.electrical_frequencies[node].frequencies
        else:
            raise ValueError()

    cpdef setup_nodes(self, list nodes):
        cdef:
            Py_ssize_t i, s_rhs_idx, s_f_idx, Nsm, Neq, Nhom, Nf
            NodeInfoEntry *info
            frequency_info_t *finfo_ptr = NULL
            FrequencyContainer ficnt

        self.nodes = {n.full_name : n for n in nodes}
        self.nodes_idx = {n.full_name : i for i, n in enumerate(nodes)}
        self.num_nodes = len(self.nodes)

        self._c_node_info = <NodeInfoEntry*> calloc(self.num_nodes, sizeof(NodeInfoEntry))
        if not self._c_node_info:
            raise MemoryError()

        Nhom = self.nhoms
        s_rhs_idx = 0 # total number of states in the matrix so far
        s_f_idx = 0 # Number of frequency submatrices so far

        for i, n in enumerate(self.nodes.values()):
            self.get_node_matrix_params(n, &Neq, &Nsm, &finfo_ptr)
            info = &self._c_node_info[i]
            info.index = i
            info.rhs_index = s_rhs_idx
            info.freq_index = s_f_idx
            info.nfreqs = Nsm
            info.nhoms = Neq
            info.frequencies = finfo_ptr

            s_rhs_idx += Neq * Nsm  # Track how many equations we are going through
            s_f_idx   += Nsm        # keep track of how many frequencies*nodes

    @contextlib.contextmanager
    def component_edge_fill(self, comp, edgestr, f1, f2, conjugate = False):
        """
        Returns a matrix for the submatrix an element has requested
        for different connections it needs. The key is::

            (element, connection_name, ifreq, ofreq)

        This is a context manager, to be used like with sim.component_edge_fill(key) as mat::

            mat[:] = computations

        Parameters
        ----------
        element : finesse.component.Connector
            The object reference that created the requests.
        connection_name : str
            String name given to the connection.
        ifreq : finesse.Frequency
            Incoming frequency.
        ofreq : finesse.Frequency
            Outgoing frequency.

        Returns
        -------
        matrix
        """
        #the two-stage nature of this will make some fill checks and hooks
        #much more safe (and powerfull)

        #this will be helpful for on-demand filling and will also help improve
        #make DC simulation of squeezers work (because post-fill transformations
        #will be needed)
        key = (comp, edgestr, f1.index, f2.index)
        mat = self._submatrices[key]
        yield mat
        #check things, or transform things here
        if conjugate:
            mat[:].imag *= -1
        return

    @contextlib.contextmanager
    def component_edge_fill3(self, owner_idx, conn_idx, f1_index, f2_index):
        mat = self._submatrices[(owner_idx, conn_idx, f1_index, f2_index)]
        yield mat
        return

    cdef assign_submatrices(self, connector_workspaces):
        """An important function. This takes all the connector workspaces - i.e. model elements
        that have requested some type of connection in the model - and ensures they have the
        correct submatrix allocated to them in for this solver.
        """
        cdef:
            NodeInfoEntry node_inf
            Frequency ifreq, ofreq
            bint couples_f, is_freq_gen
            ConnectorWorkspace ws

        if self._submatrices:
            raise Exception("Submatrices already assigned")

        self._submatrices  = {}

        from finesse.components import FrequencyGenerator, Space

        # Add in the diagonal elements of the matrix
        for n, node_info_idx in self.nodes_idx.items():
            node_inf = self._c_node_info[node_info_idx]
            Nsm = node_inf.nfreqs
            Neq = node_inf.nhoms

            for freq in range(Nsm):
                fidx = self.findex(n, freq)  # Get index for this submatrix
                self._M.add_diagonal_elements(
                    Neq, fidx, f"I,node={n},f={freq},fidx={fidx},Neq={Neq}"
                )

        id_owner = -1
        # for everything that needs to fill the matrix...
        for ws in connector_workspaces:
            owner = ws.owner
            id_owner += 1
            ws.owner_id = id_owner # store the owner index

            idx_connection = -1
            N_connections = owner._registered_connections
            SMs = []

            # For each connection this element wants...
            for name in owner._registered_connections:
                #print(name)
                idx_connection += 1

                if self.is_signal_matrix:
                    ws_conn = ws.signal.connections
                    conn_settings = ws.signal.connection_settings
                else:
                    ws_conn = ws.carrier.connections
                    conn_settings = ws.carrier.connection_settings

                # convert weak ref (input, output)
                nio = []
                for _ in owner._registered_connections[name]:
                    nio.append(owner.nodes[_])

                enabled_check = owner._enabled_checks.get(name, None)
                if enabled_check:
                    enabled_check = enabled_check()
                else:
                    enabled_check = True

                nio = tuple(nio)

                # If we are a carrier matrix only compute optics, no AC electronics or mechanics
                if not self.is_signal_matrix:
                    if (nio[0].type is not NodeType.OPTICAL
                        or nio[1].type is not NodeType.OPTICAL) or not enabled_check:
                        #print("excluded", name)
                        continue
                else:
                    # elec and mech nodes from a connection are not all necessarily modelled
                    # check if they are in the node list for this simulation
                    if (not (nio[0].full_name in self.nodes_idx and nio[1].full_name in self.nodes_idx)) or not enabled_check:
                        # If this connection hasn't been allocated then we set the
                        # matrix view array which is stored in the workspace connections info
                        # to None, so that fill methods can quickly check if they should
                        # touch it or not
                        idx_connection -= 1 # reduce connection idx count as we aren't allocating it now
                        if not hasattr(ws_conn, name):
                            setattr(ws_conn, name, None)
                        setattr(ws_conn, name + "_idx", -1)
                        continue

                dim = 0 # Dimension of the frequency coupling matrix
                # If we are not using a specialist connections object then
                # we need to add something to the generic Connections
                Nfi = self._c_node_info[self.node_id(nio[0])].nfreqs
                Nfo = self._c_node_info[self.node_id(nio[1])].nfreqs

                #print("!!!", owner, nio, Nfi, Nfo)
                ifreqs = self.get_node_frequencies(nio[0])
                ofreqs = self.get_node_frequencies(nio[1])
                #print("   in", nio[0], ifreqs, "\n   out", nio[1], ofreqs)

                # Frequency generators might couple fields, they might not.
                # so by default we set them to.
                is_freq_gen = isinstance(owner, FrequencyGenerator)
                # If the node type is different then we also are probably
                # coupling multiple frequencies together. For examaple,
                # Rad pressure, couples all sideband/carrier beats into
                # a single force state
                if hasattr(owner, '_couples_frequency'):
                    does_f_couple = owner._couples_frequency
                else:
                    does_f_couple = None

                couples_f = (
                    (
                        is_freq_gen
                        or (nio[0].type != nio[1].type)
                        or does_f_couple is not None
                    )
                    # Only if one of the nodes is optical do we have multiple
                    # frequencies to couple into
                    and (nio[0].type == NodeType.OPTICAL or nio[1].type == NodeType.OPTICAL)
                )
                #print(f"   is_freq_gen={is_freq_gen} couples_f={couples_f}")

                if not hasattr(ws_conn, name):
                    #print("NOT DEFINED", )
                    if couples_f:
                        # We will need a 2D array of submatrices to describe how multiple
                        # elements will couple together
                        setattr(ws_conn, name, np.empty((Nfi, Nfo) , dtype=object))
                        dim = 2
                    else:
                        # if not, just a diagonal
                        setattr(ws_conn, name, np.empty(Nfi, dtype=object))
                        dim = 1
                else:
                    #print("DEFINED")
                    # If a specialist object already exists lets probe it's shape
                    # as that will describe what can actually be coupled or not
                    dim = getattr(ws_conn, name).dimensions

                # Loop over all the frequencies we can couple between and add
                # submatrixes to the overall model
                for ifreq in ifreqs:
                    for ofreq in ofreqs:
                        #print("   &&& TRY ", ifreq, ofreq, does_f_couple)
                        # For each input and output frequency check if our
                        # element wants to couple them at this
                        if (
                            couples_f
                            and (does_f_couple is not None and not does_f_couple(ws, name, ifreq, ofreq))
                        ):
                            continue
                        elif not couples_f and ifreq.index != ofreq.index:
                            # If it doesn't couple frequencies and the
                            # frequencies are different then ignore
                            continue

                        #print("   &&& ACCEPT ", ifreq, ofreq)

                        iodx = []  # submatrix indices
                        tags = []  # descriptive naming tags for submatrix key
                        key_name = re.sub(r"^[^.]*\.", "", name)
                        key_name = re.sub(r">[^.]*\.", ">", key_name)
                        key = [id_owner, idx_connection]

                        # Get simulation unique indices for submatrix
                        # position. How we get these depends on the type of
                        # the nodes involved
                        for freq, node in zip((ifreq, ofreq), nio):
                            iodx.append(self.findex(node, freq.index))
                            tags.append(freq.name)
                            key.append(freq.index)

                        assert len(iodx) == 2
                        assert len(key) == 4

                        # Here we determined whether to conjugate fill a submatrix view or not
                        conjugate_fill = False
                        if self.is_signal_matrix:
                            if nio[0].type == nio[1].type == NodeType.OPTICAL:
                                # Opt-2-Opt lower sideband is conjugated
                                if ifreq.audio_order < 0 and ofreq.audio_order < 0:
                                    conjugate_fill = True
                            elif nio[0].type == NodeType.OPTICAL and ifreq.audio_order < 0:
                                # Opt-2-? lower sideband is conjugated
                                conjugate_fill = True
                            elif nio[1].type == NodeType.OPTICAL and ofreq.audio_order < 0:
                                # ?-2-Opt lower sideband is conjugated
                                conjugate_fill = True

                        if tuple(key) not in self._submatrices:
                            smname = "{}__{}__{}".format(name, *tags)

                            #print("Requesting:", key)

                            # Then we get a view of the underlying matrix which we set the values
                            # with. Store one for each frequency. By requesting this view we are
                            # telling the matrix that these elements should be non-zero in the
                            # model.
                            setting = conn_settings.get(name)
                            if setting is None:
                                # default to using full matrix if nothing set
                                setting = ConnectionSetting.MATRIX

                            if setting == ConnectionSetting.DIAGONAL:
                                #print("!!!D", owner, name, self.is_signal_matrix)
                                SM = self._M.get_sub_diagonal_view(*iodx, smname, conjugate_fill)
                            elif setting == ConnectionSetting.MATRIX:
                                #print("!!!M", owner, name, self.is_signal_matrix)
                                SM = self._M.get_sub_matrix_view(*iodx, smname, conjugate_fill)
                            elif setting == ConnectionSetting.DISABLED:
                                #print("!!!DIS", owner, name, self.is_signal_matrix)
                                SM = None
                            else:
                                raise Exception(f"Unhandled setting {setting}")
                            #print("!@#", owner, name, self.is_signal_matrix, dim)
                            try:
                                if dim == 1:
                                    getattr(ws_conn, name)[ifreq.index] = SM
                                elif dim == 2:
                                    getattr(ws_conn, name)[ifreq.index, ofreq.index] = SM
                                else:
                                    raise Exception(f"Unhandled dimension size {dim}")
                            except IndexError:
                                raise IndexError(f"Error setting submatrix to connection {name} in {owner}. "
                                                  "Size of array of submatricies wrong, number of frequencies "
                                                  "assumed probably incorrect.")

                            setattr(ws_conn, name + "_idx", idx_connection)
                            self._submatrices[tuple(key)] = SM
                            SMs.append(SM)
                        else:
                            # Check if we've just requested the same submatrix.
                            sm = self._submatrices[tuple(key)]
                            if sm.from_idx != iodx[0] or sm.to_idx != iodx[1]:
                                raise Exception(
                                    "Requested submatrix has already been requested,"
                                    "but new one has different indices"
                                )
                            else:
                                continue
        #print("done")

    cdef assign_noise_submatrices(self, connector_workspaces):
        import itertools
        cdef NodeInfoEntry node_inf
        cdef Frequency ifreq, ofreq
        cdef ConnectorWorkspace ws
        cdef int i

        self._noise_submatrices  = {}

        for noise_type, sources in self.noise_sources.items():
            M = self._noise_matrices[noise_type]
            self._noise_submatrices[noise_type] = {}

            # Add in the diagonal elements of the matrices
            for n, node_info_idx in self.nodes_idx.items():
                node_inf = self._c_node_info[node_info_idx]
                Nsm = node_inf.nfreqs
                Neq = node_inf.nhoms
                for freq in range(Nsm):
                    fidx = self.findex(n, freq)  # Get index for this submatrix
                    mat = M.add_diagonal_elements(
                        Neq, fidx, f"I,node={n},f={freq},fidx={fidx},Neq={Neq}"
                    )
                    self._noise_submatrices[noise_type][fidx] = mat

            for comp, nodes in sources:
                ws = None
                for _ws in self.workspaces.to_noise_refill:
                    if _ws.owner is comp:
                        ws = _ws
                        break
                if ws is None:
                        raise Exception("Noise source not registered")
                for name, node in nodes:
                    freqs = self.get_node_frequencies(node)

                    if hasattr(comp, '_couples_noise'):
                        couples_noise = comp._couples_noise
                    else:
                        couples_noise = None

                    # Loop over all the noise sidebands we can couple between and add
                    # submatrixes to the overall model
                    for ifreq, ofreq in itertools.product(freqs, freqs):
                        if couples_noise is None:
                            if ifreq.index != ofreq.index:
                                continue
                        elif not couples_noise(ws, node, noise_type, ifreq, ofreq):
                            continue

                        iodx = []  # submatrix indices
                        tags = []  # descriptive naming tags for submatrix key
                        key = [ws.owner_id, self.node_id(node)]

                        # Get simulation unique indices for submatrix position.
                        for freq in [ifreq, ofreq]:
                            iodx.append(self.findex(node, freq.index))
                            tags.append(freq.name)
                            key.append(freq.index)

                        assert len(iodx) == 2
                        assert len(key) == 4

                        # Here we determined whether to conjugate fill a submatrix view or not
                        conjugate_fill = False
                        if node.type == NodeType.OPTICAL:
                            # Opt-2-Opt lower sideband is conjugated
                            if ifreq.audio_order < 0 and ofreq.audio_order < 0:
                                conjugate_fill = True

                        if tuple(key) not in self._noise_submatrices[noise_type]:
                            smname = "{}__{}__{}".format(name, *tags)

                            if ifreq == ofreq:
                                SM = self._noise_submatrices[noise_type][self.findex(node, ifreq.index)]
                            else:
                                SM = M.get_sub_matrix_view(*iodx, smname, conjugate_fill)
                            getattr(ws.signal.noise_sources, name)[ifreq.index, ofreq.index] = SM

                            self._noise_submatrices[noise_type][tuple(key)] = SM
                        else:
                            # Check if we've just requested the same submatrix.
                            sm = self._noise_submatrices[noise_type][tuple(key)]
                            if sm.from_idx != iodx[0] or sm.to_idx != iodx[1]:
                                raise Exception(
                                    "Requested submatrix has already been requested,"
                                    "but new one has different indices"
                                )
                            else:
                                continue

        if NoiseType.QUANTUM in self.noise_sources:
            M = self._noise_matrices[NoiseType.QUANTUM]
            for ws in connector_workspaces:
                for i in range(ws.input_noise.num_nodes):
                    n = ws.input_noise.node_info[i].idx
                    node_inf = self._c_node_info[n]
                    Nsm = node_inf.nfreqs
                    for freq in range(Nsm):
                        fidx = self.findex_fast(n, freq)  # Get index for this submatrix
                        ws.input_noise.nodes[i, freq] = self._noise_submatrices[NoiseType.QUANTUM][fidx]
                for i in range(ws.output_noise.num_nodes):
                    n = ws.output_noise.node_info[i].idx
                    node_inf = self._c_node_info[n]
                    Nsm = node_inf.nfreqs
                    for freq in range(Nsm):
                        fidx = self.findex_fast(n, freq)  # Get index for this submatrix
                        ws.output_noise.nodes[i, freq] = self._noise_submatrices[NoiseType.QUANTUM][fidx]

    cdef add_noise_matrix(self, object key):
        raise NotImplementedError

    cdef add_rhs(self, unicode key):
        raise NotImplementedError

    cdef update_frequency_info(self):
        self.optical_frequencies.update_frequency_info()
        if self.is_signal_matrix:
            for i in range(len(self.unique_elec_mech_fcnts)):
                fcnt = (<FrequencyContainer>self.unique_elec_mech_fcnts[i]).update_frequency_info()

    cdef factor(self):
        raise NotImplementedError()

    cdef refactor(self):
        raise NotImplementedError()

    cdef solve(self):
        raise NotImplementedError()

    cdef solve_noises(self):
        raise NotImplementedError()

    cdef initial_fill(self):
        cdef Py_ssize_t i
        cdef ConnectorWorkspace ws
        cdef ConnectorMatrixSimulationInfo cmsinfo
        cdef fill_list_t *fill_list

        self.optical_frequencies.initialise_frequency_info()
        if self.is_signal_matrix:
            for i in range(len(self.unique_elec_mech_fcnts)):
                (<FrequencyContainer>self.unique_elec_mech_fcnts[i]).initialise_frequency_info()

        for ws in self.workspaces.to_initial_fill:
            ws.update_parameter_values()
            if self.is_signal_matrix:
                cmsinfo = ws.signal
            else:
                cmsinfo = ws.carrier
            fill_list = &cmsinfo.matrix_fills

            for i in range(fill_list.size):
                if fill_list.infos[i].fn_c:
                    fill_list.infos[i].fn_c(ws)
                elif fill_list.infos[i].fn_py:
                    (<object>fill_list.infos[i].fn_py).__call__(ws)

    cdef refill(self):
        cdef Py_ssize_t i, j
        cdef ConnectorWorkspace ws
        cdef ConnectorMatrixSimulationInfo cmsinfo
        cdef fill_list_t *fill_list

        if self.any_frequencies_changing:
            self.update_frequency_info()

        for i in range(self.workspaces.num_to_refill):
            ws = <ConnectorWorkspace>self.workspaces.ptr_to_refill[i]
            # TODO (sjr) Probably don't need this update call for now
            #            (see start of self.run method)
            ws.update_parameter_values()
            if self.is_signal_matrix:
                fill_list = &ws.signal.matrix_fills
            else:
                fill_list = &ws.carrier.matrix_fills

            for j in range(fill_list.size):
                if fill_list.infos[j].refill or self.forced_refill:
                    if fill_list.infos[j].fn_c:
                        fill_list.infos[j].fn_c(ws)
                    elif fill_list.infos[j].fn_py:
                        (<object>fill_list.infos[j].fn_py).__call__(ws)

        # As we have changed the matrix elements we need to refactor
        self.refactor()

    cdef fill_rhs(self):
        cdef ConnectorWorkspace ws
        cdef ConnectorMatrixSimulationInfo sys_ws
        for ws in self.workspaces.to_rhs_refill:
            ws.update_parameter_values()

            if self.is_signal_matrix:
                sys_ws = ws.signal
            else:
                sys_ws = ws.carrier

            if sys_ws.fn_rhs_c is not None:
                sys_ws.fn_rhs_c.func(ws)
            elif sys_ws.fn_rhs_py is not None:
                sys_ws.fn_rhs_py(ws)

    cdef fill_noise_inputs(self):
        cdef ConnectorWorkspace ws
        for ws in self.workspaces.to_noise_input_refill:
            if NoiseType.QUANTUM in self.noise_sources:
                if ws.signal.fn_quantum_noise_input_c is not None:
                    ws.signal.fn_quantum_noise_input_c.func(ws)
                elif ws.signal.fn_quantum_noise_input_py is not None:
                    ws.signal.fn_quantum_noise_input_py(ws)

        for ws in self.workspaces.to_noise_refill:
            if NoiseType.QUANTUM in self.noise_sources:
                if ws.signal.fn_quantum_noise_c is not None:
                    ws.signal.fn_quantum_noise_c.func(ws)
                elif ws.signal.fn_quantum_noise_py is not None:
                    ws.signal.fn_quantum_noise_py(ws)

    cdef construct(self):
        """This is called when workspaces and submatrices have been setup. Calling
        construct should now go and allocate the memory for the matrix and RHS.

        This method should be overwritten by an inheriting solver class with
        specfics of the solving technique.
        """
        raise NotImplementedError()

    cdef initial_run(self):
        """Once a solver has been constructed it will most likely need to be initially
        filled and ran. Some sparse solvers for example must do a full factor first, then
        can perform faster refactors.

        This method should be overwritten by an inheriting solver class with
        specfics of the solving technique.
        """
        raise NotImplementedError()

    cpdef run(self):
        """Executes the simulation for model in its current state.

        Takes the following steps to compute an output:
         * If self.manual_rhs:
            * Clears the RHS vector
            * Fills the RHS vector
         * Fills the matrix
         * Solves
        """
        if not self.manual_rhs:
            self.clear_rhs()
            self.fill_rhs()
        self.refill()
        self.solve()

    @property
    def M(self):
        """
        A weak reference to the underlying Matrix object.

        .. note::

            References to the Matrix should not be kept.

        :getter: Returns a weak reference to the underlying matrix (read-only).
        """
        return weakref.ref(self._M)

    def print_matrix(self):
        self._M.print_matrix()

    cpdef clear_rhs(self):
        self._M.clear_rhs()

    def set_source(self, object node, int freq_idx, int hom_idx, complex value):
        self._M.set_rhs(self.field_fast(self.node_id(node), freq_idx, hom_idx), value)

    cdef void set_source_fast(self, Py_ssize_t node_id, Py_ssize_t freq_idx, Py_ssize_t hom_idx, complex_t value, unsigned rhs_index=0):
        self._M.set_rhs(self.field_fast(node_id, freq_idx, hom_idx), value, rhs_index)

    cdef void set_source_fast_2(self, Py_ssize_t rhs_idx, complex_t value):
        self._M.set_rhs(rhs_idx, value, 0)

    cdef void set_source_fast_3(self, Py_ssize_t rhs_idx, complex_t value, unsigned rhs_index):
        self._M.set_rhs(rhs_idx, value, rhs_index)

    cpdef Py_ssize_t findex(self, object node, Py_ssize_t freq):
        """
        Returns simulation unique index for a given frequency at this node.
        Used to refer to submatrices of HOMs in the interferometer matrix.

        Parameters
        ----------
        node : :class:`.Node`
            Node object to get the index of.
        freq : int
            Frequency index.

        Returns
        -------
        index : int
            Index of the `node` for a given frequency.
        """
        return self.findex_fast(self.node_id(node), freq)

    cdef Py_ssize_t findex_fast(self, Py_ssize_t node_id, Py_ssize_t freq) nogil:
        cdef:
            NodeInfoEntry ni = self._c_node_info[node_id]
            Py_ssize_t freq_idx = ni.freq_index

        return freq_idx + freq

    cpdef Py_ssize_t field(self, object node, Py_ssize_t freq=0, Py_ssize_t hom=0):
        """
        Returns simulation unique index of a field at a particular frequency
        index at this node.

        Parameters
        ----------
        node : :class:`.Node`
            Node object to get the index of.
        freq : int
            Frequency index.
        hom : int, optional
            Higher Order Mode index, defaults to zero.
        """
        return self.field_fast(self.node_id(node), freq, hom)

    cdef Py_ssize_t field_fast(self, Py_ssize_t node_id, Py_ssize_t freq=0, Py_ssize_t hom=0) nogil:
        cdef:
            NodeInfoEntry ni = self._c_node_info[node_id]
            Py_ssize_t Nh = ni.nhoms
            Py_ssize_t rhs_idx = ni.rhs_index
        return rhs_idx + freq * Nh + hom

    cdef inline Py_ssize_t field_fast_2(
        self,
        Py_ssize_t node_rhs_idx,
        Py_ssize_t num_hom,
        Py_ssize_t freq,
        Py_ssize_t hom) nogil:
        """Inlined function to return field index fast."""
        return node_rhs_idx + freq * num_hom + hom

    cpdef complex_t get_out(self, object node, Py_ssize_t freq=0, Py_ssize_t hom=0):
        return self.get_out_fast(self.node_id(node), freq, hom)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.initializedcheck(False)
    cdef complex_t get_out_fast(self, Py_ssize_t node_id, Py_ssize_t freq=0, Py_ssize_t hom=0) nogil:
        cdef Py_ssize_t field_idx = self.field_fast(node_id, freq, hom)

        if self.out_view is None:
            return COMPLEX_0

        return self.out_view[field_idx]

    cpdef Py_ssize_t node_id(self, object node):
        if type(node) is str:
            return self.nodes_idx[node]
        else:
            return self.nodes_idx[node.full_name]

    cpdef get_node_info(self, node):
        """For a given node (object or name) the key parameters
        for where this node is represented in the matrix of linear
        equations is returned in a NodeInfoEntry object.
        """
        cdef int i
        if type(node) is str:
            i = self.nodes_idx[node]
        else:
            i = self.nodes_idx[node.full_name]

        cdef NodeInfoEntry ni = self._c_node_info[i]
        return {
            "index": ni.index,
            "rhs_index": ni.rhs_index,
            "freq_index": ni.freq_index,
            "nfreqs": ni.nfreqs,
            "nhoms": ni.nhoms,
        }

    def get_frequency_object(self, frequency, node):
        """Get a :class:`.Frequency` object corresponding to a numerical or symbolic value.
        Returns none if nothing has been found.

        Parameters
        ----------
        f : number or :class:`.Symbol`
            Frequency to search for in this simulation.

        Returns
        -------
        :class:`.Frequency`
            The frequency object.
        """
        from finesse.symbols import Symbol

        if node.type == NodeType.OPTICAL:
            frequencies = self.optical_frequencies.frequencies
        elif node.type == NodeType.MECHANICAL:
            if not self.is_signal_matrix:
                return None
            frequencies = self.mechanical_frequencies[node].frequencies
        elif node.type == NodeType.ELECTRICAL:
            if not self.is_signal_matrix:
                return None
            frequencies = self.electrical_frequencies[node].frequencies

        if isinstance(frequency, Symbol):
            if frequency.is_changing:
                # if it's tunable we want to look for the symbol that is just this
                # lasers frequency, as it will be changing
                for f in frequencies:
                    if f.symbol == frequency:
                        return f

        f_value = float(frequency)
        # otherwise do some value comparisons
        for f in frequencies:
            if np.isclose(float(f.f), f_value, atol=1e-15, rtol=1e-15):
                return f

        return None


cdef class CarrierSignalMatrixSimulation:
    """A matrix based simulation which models the DC build of optical
    fields (carriers) and optionally a signal simulation, which models
    small AC oscillations throughout the system around the DC operating
    point.

    This simualations class contains two solvers, one for the carrier
    optical fields, and another for the signal simulations consisting of
    optical, electrical, and mechanical.
    """
    def __init__(self, model, unicode name, bint needs_matrix=True):
        self.model = model
        self.name = name
        self.trace = NULL
        self.mismatched_node_couplings = ()
        self.contingent_trace_forests = {}
        self.needs_reflag_changing_q = False
        self.do_matrix_solving = needs_matrix

    def __dealloc__(self):
        if self.trace:
            free(self.trace)

    def build(self):
        from finesse.simulations.KLU import KLUMatrixSystemSolver

        if self.model.fsig.f.is_changing and self.model.fsig.f.value is None:
            #raise Exception(f"Signal frequency is being changed but an initial value has not been set. Set initial value for `fsig` before running.")
            self.model.fsig.f.value = 1

        # Keep a list of all the things changing in the simulation
        self.changing_parameters = set(
            p for el in self.model.elements.values() for p in el.parameters if p.is_changing
        )
        self.tunable_parameters = set(
            p for p in self.changing_parameters if p.is_tunable
        )
        self.compute_signals = self.model.fsig.f.value is not None
        self.is_modal = self.model.is_modal
        self.initialise_model_data()
        self.initialise_sim_config_data()

        # First we sort out the two matrix simulations as required by the model
        cf = self.generate_carrier_frequencies()
        onodes = self.model.optical_nodes
        self.carrier = KLUMatrixSystemSolver("carrier", onodes, cf, None, None, self.model_data.num_HOMs, False, self.model.force_refill)

        if self.compute_signals:
            nodes = onodes.copy()
            nodes.extend(self.model.get_active_elec_mech_nodes())
            sf, mf, ef = self.generate_signal_frequencies(nodes, cf)
            self.signal = KLUMatrixSystemSolver("signal", nodes, sf, mf, ef, self.model_data.num_HOMs, True, self.model.force_refill)
            self.initialise_noise_matrices()

        self.initialise_trace_forest(onodes)
        self.initialise_workspaces()
        self.update_all_parameter_values()

        if self.compute_signals:
            self.initialise_noise_sources()
            self.initialise_noise_selection_vectors()

        if self.is_modal:
            # compute all the initial:
            #     - scattering matrices
            #     - space Gouy phases
            #     - laser tem Gouy phases
            if self.do_matrix_solving:
                self.compute_knm_matrices()
            self.set_gouy_phases()
            # ... then determine which beam parameters will be changing
            # so that only the items from above which change get
            # re-computed on subsequent calls to their respective functions
            self._determine_changing_beam_params()

        if self.do_matrix_solving:
            self.carrier.assign_submatrices(self.workspaces)
            self.carrier.construct()
            self.carrier.initial_run()

            if self.signal:
                self.signal.assign_submatrices(self.workspaces)
                self.signal.assign_noise_submatrices(self.workspaces)
                self.signal.construct()
                self.signal.initial_run()

        self.setup_output_workspaces()

    def unbuild(self):
        pass
        #self._unbuild()
        #self._clear_workspace()

        # Code below can be used in debug mode to determine if anyone is keeping any
        # references to this matrix object, meaning its memory can't be freed.
        # This takes ~20ms to do so makes a difference for quick models. Maybe we need
        # a debug mode

        #_ref = self._M
        #self._M = None

        # refs = gc.get_referrers(_ref)
        # Nref = len(refs)
        # if Nref > 0:
        #     LOGGER.warn(
        #         f"Something other than the Simulation object (N={Nref}) is keeping"
        #         f" a reference to the matrix object {repr(self._M)}."
        #         " Could lead to excessive memory usage if not released."
        #     )
        #     for _ in refs:
        #         LOGGER.warn(f" - {repr(_)}")
        #del _ref


    cdef void update_cavities(self):
        cdef CavityWorkspace ws
        for ws in self.cavity_workspaces.values():
            ws.update()

    cdef void compute_knm_matrices(self):
        cdef KnmConnectorWorkspace ws
        for ws in self.to_scatter_matrix_compute:
            ws.update_changing_knm_workspaces()
            ws.compute_scattering_matrices()
            # TODO (sjr) Probably want a flag to check if quantum noise calcs
            #            being performed and to only do this call if so
            ws.compute_knm_losses()

    cdef int set_gouy_phases(self) except -1:
        cdef ConnectorWorkspace ws
        cdef int rtn

        for ws in self.gouy_phase_workspaces:
            if ws.fn_gouy_c is not None:
                rtn = ws.fn_gouy_c.func(ws)
            elif ws.fn_gouy_py is not None:
                rtn = ws.fn_gouy_py(ws)
            if rtn:
                return rtn

        return 0

    cpdef int modal_update(self) except -1:
        """Updates HOM related dependencies / properties of the model.

        These updates are as follows:

         * Execute a beam trace on the changing trace trees
         * Computes the changing scattering matrices
         * Calculates the Gouy phase of Spaces and Laser power coefficients

        Returns
        -------
        exit_code : int
            A value of ``0`` if successful, a value of ``1`` if an error
            was raised during the beam tracing.
        """
        cdef bint beam_trace_valid = True

        self.update_cavities()

        if self.retrace:
            beam_trace_valid = self.trace_beam()
            if not beam_trace_valid:
                return 1

        # compute the changing scattering matrices
        if self.do_matrix_solving:
            self.compute_knm_matrices()

        # update the changing Gouy phases at spaces
        # and TEM Gouy phases at lasers
        return self.set_gouy_phases()

    cdef void _determine_changing_beam_params(
        self, TraceForest forest=None, bint set_tree_node_ids=True,
    ):
        cdef:
            Py_ssize_t i
            Py_ssize_t num_nodes = self.carrier.num_nodes
            ConnectorWorkspace ws
            KnmConnectorWorkspace kws

        # Re-set all beam parameter changing flags to false initially
        for i in range(num_nodes):
            self.trace[i].is_changing = False

        if self.retrace:
            LOGGER.info("Flagging changing beam parameters.")
            # Prepare the forest for simulation by setting all the node_id attributes
            # and flag the corresponding self.trace entries as changing
            self._setup_trace_forest(forest, set_tree_node_ids)

        # Now tell each knm workspace whether it is changing or not
        # so that only changing scattering matrices get recomputed
        # from here on
        for kws in self.to_scatter_matrix_compute:
            kws.flag_changing_knm_workspaces()

    def is_component_in_mismatch_couplings(self, comp):
        """Determines whether the connector `comp` is associated with any
        of the node couplings in the stored potential mismatch couplings.

        .. note::

            This method can be replaced if connectors eventually use more
            granular refill flags --- i.e. per coupling refill flags. Then
            the check for refilling that coupling can simply include the
            condition ``(from_node, to_node) in sim.mismatched_node_couplings``.
        """
        return any(node.component is comp for node, _ in self.mismatched_node_couplings)

    cdef void _setup_trace_forest(self, TraceForest forest=None, bint set_tree_node_ids=True):
        cdef:
            Py_ssize_t tree_idx
            TraceTree tree

        if forest is None:
            forest = self.trace_forest

        for tree_idx in range(forest.N_trees):
            tree = forest.forest[tree_idx]
            self._setup_single_trace_tree(tree, set_tree_node_ids)

    cdef void _setup_single_trace_tree(self, TraceTree tree, bint set_tree_node_ids=True):
        cdef:
            TraceTree ltree = tree.left
            TraceTree rtree = tree.right

        # Only ever need to do this once, so avoid repeating when reflagging
        # changing beam params after exiting unstable cavity regions
        if set_tree_node_ids:
            tree.node_id = self.carrier.node_id(tree.node)
            tree.opp_node_id = self.carrier.node_id(tree.node.opposite)

        self.trace[tree.node_id].is_changing = True
        self.trace[tree.opp_node_id].is_changing = True

        if ltree is not None:
            self._setup_single_trace_tree(ltree)
        if rtree is not None:
            self._setup_single_trace_tree(rtree)

    cdef tuple _find_new_unstable_cavities(self):
        cdef:
            Py_ssize_t tree_idx
            TraceTree tree

            CavityWorkspace cav_ws
            bint source_is_cav
            double gx, gy

            list ch_unstable_cavities = []

        for tree_idx in range(self.trace_forest.N_trees):
            tree = self.trace_forest.forest[tree_idx]

            if tree.is_source:
                cav_ws = self.cavity_workspaces.get(tree.dependency)
                source_is_cav = cav_ws is not None
                if source_is_cav: # Tree is an internal cavity tree
                    # The geometrically changing cavity has become unstable
                    # so inform that this is the case
                    if not cav_ws.is_stable:
                        ch_unstable_cavities.append(cav_ws.owner)

                        gx = cav_ws.gx
                        gy = cav_ws.gy
                        if float_eq(gx, gy):
                            LOGGER.warning(
                                "Cavity %s is unstable with g = %s",
                                tree.dependency.name,
                                gx,
                            )
                        else:
                            LOGGER.warning(
                                "Cavity %s is unstable with gx = %s, gy = %s",
                                tree.dependency.name,
                                gx,
                                gy,
                            )

        # Return tuple of the unstable cavities sorted by name so that
        # all permutations of the same combination of cavities give same
        # tuple --- important for look-ups in contingent_trace_forests dict
        return tuple(sorted(ch_unstable_cavities, key=lambda x: x.name))

    cdef TraceForest _initialise_contingent_forest(self, tuple unstable_cavities):
        cdef TraceForest contingent_forest = TraceForest(self.model)
        cdef TraceForest model_trace_forest = self.model.trace_forest
        cdef list order = model_trace_forest.dependencies.copy()
        for uc in unstable_cavities:
            order.remove(uc)

        # If there are no dependencies left after disabling the
        # unstable cavities then a beam trace cannot be performed
        # at this data point so no forest can be planted
        if not order:
            LOGGER.warning(
                "Cannot build a contingent trace forest as the simulation "
                "is in a regime with no stable cavities nor Gauss objects."
            )
            return None

        contingent_forest.symmetric = self.trace_forest.symmetric
        contingent_forest.plant(order)

        self._determine_changing_beam_params(contingent_forest)

        return contingent_forest

    @cython.initializedcheck(False)
    cdef void _propagate_trace(self, TraceTree tree, bint symmetric):
        cdef:
            TraceTree ltree = tree.left
            TraceTree rtree = tree.right

            const NodeBeamParam* q1 = &self.trace[tree.node_id]
            complex_t qx1 = q1.qx.q
            complex_t qy1 = q1.qy.q
            complex_t qx2, qy2

        if ltree is not None:
            # For non-symmetric traces we have some special checks
            # to do on trees which couldn't be reached from the
            # other dependency trees. Note these are only performed
            # on the left tree; see TraceForest._add_backwards_nonsymm_trees
            # for details.
            if symmetric or (not tree.do_nonsymm_reverse and not tree.do_inv_transform):
                qx2 = transform_q(tree.left_abcd_x, qx1, tree.nr, ltree.nr)
                qy2 = transform_q(tree.left_abcd_y, qy1, tree.nr, ltree.nr)
            elif tree.do_inv_transform:
                # Can't reach tree directly but there is a coupling from ltree.node
                # to tree.node so apply the inverse abcd law to get correct q
                qx2 = inv_transform_q(tree.left_abcd_x, qx1, tree.nr, ltree.nr)
                qy2 = inv_transform_q(tree.left_abcd_y, qy1, tree.nr, ltree.nr)
            else:
                # Really is no way to get to the node (no coupling from ltree.node to
                # tree.node) so only option now is to reverse q for ltree node entry
                qx2 = -conj(qx1)
                qy2 = -conj(qy1)

            self.trace[ltree.node_id].qx.q = qx2
            self.trace[ltree.node_id].qy.q = qy2
            if symmetric:
                self.trace[ltree.opp_node_id].qx.q = -conj(qx2)
                self.trace[ltree.opp_node_id].qy.q = -conj(qy2)

            self._propagate_trace(ltree, symmetric)

        if rtree is not None:
            qx2 = transform_q(tree.right_abcd_x, qx1, tree.nr, rtree.nr)
            qy2 = transform_q(tree.right_abcd_y, qy1, tree.nr, rtree.nr)

            self.trace[rtree.node_id].qx.q = qx2
            self.trace[rtree.node_id].qy.q = qy2
            if symmetric:
                self.trace[rtree.opp_node_id].qx.q = -conj(qx2)
                self.trace[rtree.opp_node_id].qy.q = -conj(qy2)

            self._propagate_trace(rtree, symmetric)

    cdef bint trace_beam(self):
        cdef:
            TraceTree tree
            Py_ssize_t tree_idx

            CavityWorkspace cav_ws
            bint source_is_cav

            complex_t qx_src, qy_src

            # The actual trace forest which gets traced. In most circumstances
            # this will be self.trace_forest but if changing cavities enter an
            # unstable regime then this will be temporarily swapped out for the
            # contingent forest for this data point (see below).
            TraceForest trace_forest

            # Objects necessary for dealing with newly unstable cavities
            tuple ch_unstable_cavities
            TraceForest contingent_forest

        # No changing beam parameters, do nothing
        if not self.trace_forest.N_trees:
            return True

        # First we loop over the source trees and find any changing
        # cavities which have become unstable
        ch_unstable_cavities = self._find_new_unstable_cavities()

        # If we did find any newly unstable cavities then the current trace_forest
        # is invalidated at the current data point so we must build a new forest with
        # the unstable cavities disabled
        # NOTE (sjr) Don't worry about increased Python interaction in
        #            this block as this will rarely be executed anyway
        if ch_unstable_cavities:
            LOGGER.info(
                "Attempting to use a contingent trace forest due "
                "to the presence of unstable cavities"
            )
            # Look-up the combination of unstable cavities to see if a
            # contingent forest was already built from this
            contingent_forest = self.contingent_trace_forests.get(ch_unstable_cavities)

            # No previous forest built from the given combination of disabled
            # unstable cavities so need to build one here
            if contingent_forest is None:
                LOGGER.debug(
                    "For unstable cavity combination %s no cached contingent "
                    "trace forest found, now attempting to build a new one...",
                    [uc.name for uc in ch_unstable_cavities],
                )
                contingent_forest = self._initialise_contingent_forest(ch_unstable_cavities)
                # If there are no dependencies left after disabling the
                # unstable cavities then a beam trace cannot be performed
                # at this data point so inform of this on return
                if contingent_forest is None:
                    return False

                # Cache the contingent forest for this combination of unstable
                # cavities as these typically occur in blocks (or across strides)
                # of data points so we don't want to keep rebuilding the same
                # contingency forests for identical unstable cavity combos
                self.contingent_trace_forests[ch_unstable_cavities] = contingent_forest
            else:
                LOGGER.debug(
                    "For unstable cavity combination %s found and using "
                    "cached contingent trace forest:%s",
                    [uc.name for uc in ch_unstable_cavities],
                    contingent_forest
                )

            # Make sure only the correctly changing beam parameters, according
            # to self.trace_forest, get reflagged when exiting from the unstable
            # region again
            self.needs_reflag_changing_q = True

            # Use the contingent forest for this data point
            trace_forest = contingent_forest

        # Otherwise we just use the standard changing trace forest of the simulation
        else:
            trace_forest = self.trace_forest

            # If we've just exited an unstable cavity region where a contingent trace
            # forest was being used, then we need to reflag the beam parameters which
            # are changing
            if self.needs_reflag_changing_q:
                self._determine_changing_beam_params(forest=None, set_tree_node_ids=False)
                self.needs_reflag_changing_q = False

        # Now do the actual beam tracing by simply traversing the forest
        # and propagating the beam through each tree
        for tree_idx in range(trace_forest.N_trees):
            tree = trace_forest.forest[tree_idx]

            if tree.is_source:
                cav_ws = self.cavity_workspaces.get(tree.dependency)
                source_is_cav = cav_ws is not None

                if not source_is_cav: # Source tree is from a Gauss
                    # TODO (sjr) Should probably make some workspace for Gauss objects
                    #            which then uses cy_expr's for evaluating these things
                    #            if they're symbolic. But for now this will do.
                    qx_src = complex(tree.dependency.qx.q)
                    qy_src = complex(tree.dependency.qy.q)
                else: # Source tree is from a Cavity
                    qx_src = cav_ws.qx
                    qy_src = cav_ws.qy

                self.trace[tree.node_id].qx.q = qx_src
                self.trace[tree.node_id].qy.q = qy_src
                if trace_forest.symmetric:
                    self.trace[tree.opp_node_id].qx.q = -conj(qx_src)
                    self.trace[tree.opp_node_id].qy.q = -conj(qy_src)

            self._propagate_trace(tree, trace_forest.symmetric)

        return True

    cpdef run_carrier(self):
        """Runs the carrier matrix solver for the current state of the model.
        This will update all the C based structs with the current model state so
        that filling and calculations can be performed.

        Returns
        -------
        exit_code : int
            ``0`` if this was a valid run and the output can be used, ``-1`` otherwise.
        """
        # NOTE (sjr) Just updating all parameter values on each call to run for
        #            now. This may not be the most optimal thing to do, but it
        #            avoids duplicating these parameter update calls in different
        #            places (e.g. refill, compute_knm_matrices, set_gouy_phases) and
        #            should be safe in that no parameters get accidentally missed at
        #            any data point.
        # ddb - this just updates everything, even things that are not changing as
        # it acts on all the workspaces, probably not the best idea
        self.update_all_parameter_values()

        # update HOM stuff
        cdef int status

        if self.is_modal:
            status = self.modal_update()
            # immediately return if invalid beam trace region encountered
            # no need to go ahead and fill or solve as they won't be used
            if status:
                return status

        if self.do_matrix_solving:
            self.carrier.run()
        return 0

    cpdef run_signal(self):
        """Runs the signal matrix solver for the current state. This function should assume that
        a call to the `run_carrier` method has preceeded it. Many modal and parameter updates
        should happen in there already, so do not need to be repeated here.
        """
        cdef:
            frequency_info_t freq
        self.model_data.fsig = self.model.fsig.f.value
        # Probably some other preparatory stuff needs to go here in the future
        if self.do_matrix_solving:
            self.signal.run()

            # Then ask components for their noise contributions
            if self.signal.noise_sources:
                self.signal.fill_noise_inputs()
                self.signal.solve_noises()

    def setup_output_workspaces(self):
        # Once the simulations are started we can tell all the detectors to
        # prepare themselves
        self.detector_workspaces = []
        self.readout_workspaces = []

        for rd in self.model.get_elements_of_type(Readout):
            # Readouts can emulate multiple detectors, so here we
            # get a collection of them depending on what the readout
            # is doing and add them to the list
            wss = rd._get_output_workspaces(self)
            if wss is not None:
                for ws in wss: # check we can iterate over the returned workspaces
                    if not isinstance(ws, DetectorWorkspace):
                        raise TypeError(f"Readout detector ({rd}) workspace ({ws}) not a DetectorWorkspace type")
                    self.readout_workspaces.append(ws)
                    if rd.output_detectors:
                        self.detector_workspaces.append(ws)

        for det in self.model.detectors:
            ws = det._get_workspace(self)

            if not isinstance(ws, DetectorWorkspace):
                raise TypeError(f"Detector ({det}) workspace ({ws}) not a DetectorWorkspace type")

            self.detector_workspaces.append(ws)
            if self.signal and isinstance(ws.owner, NoiseDetector):
                self.signal.workspaces.noise_detectors.append(ws)

        if self.signal:
            self.signal.workspaces.detector_list_to_C()

    def __enter__(self):
        self.build()

    def __exit__(self, type_, value, traceback):
        self.unbuild()

    cdef void update_all_parameter_values(self):
        """Loops through all workspaces to update the C structs so they
        represent the current model element parameter values.
        """
        cdef:
            ElementWorkspace ews
            ConnectorWorkspace ws

            Py_ssize_t i
            # TODO (sjr) Should probably cache these or move away from
            #            lists for best performance
            Py_ssize_t Nvws = len(self.variable_workspaces)
            Py_ssize_t Ncws = len(self.workspaces)

        for i in range(Nvws):
            ews = self.variable_workspaces[i]
            ews.update_parameter_values()

        for i in range(Ncws):
            ws = self.workspaces[i]
            ws.update_parameter_values()

    cdef initialise_trace_forest(self, optical_nodes):
        cdef TraceForest model_trace_forest
        cdef double nr
        # Before we setup the workspaces some initial beam trace must be done
        # so that workspaces can initialise themselves
        if self.is_modal:
            # Make sure the model trace forest gets re-planted
            # when building a new simulation
            self.model._rebuild_trace_forest = True
            LOGGER.info(
                "Performing initial beam trace with configuration options: %s",
                self.model.sim_trace_config,
            )
            self.model.beam_trace(**self.model.sim_trace_config)
            model_trace_forest = self.model.trace_forest
            self.retrace = self.model.retrace
            self.trace = <NodeBeamParam*> calloc(len(optical_nodes), sizeof(NodeBeamParam))
            if not self.trace:
                raise MemoryError()

            for i, n in enumerate(optical_nodes):
                qx, qy = self.model.last_trace[n]
                nr = qx.nr

                # NOTE set is_changing to True for all node beam params initially
                #      so that initial scattering matrices get computed, these flags
                #      will then be updated in _determine_changing_beam_params
                self.trace[i] = NodeBeamParam(
                    beam_param(qx.q, nr, self.model_data.lambda0),
                    beam_param(qy.q, nr, self.model_data.lambda0),
                    True
                )

            if self.retrace:
                # Construct the forest of changing trace trees - it's important
                # that this is done before initialising connector workspaces as
                # they need the changing forest to be present for refill flag
                self.trace_forest = model_trace_forest.make_changing_forest()

                if self.trace_forest.N_trees:
                    LOGGER.info(
                        "Determined changing trace trees:%s", self.trace_forest
                    )

                # Get the nodes at which trees of the changing forest intersect
                # with trees of the full forest which have different trace
                # dependencies. These couplings will be potentially mismatched
                # during the simulation.
                self.mismatched_node_couplings = self.trace_forest.find_potential_mismatch_couplings(
                    model_trace_forest
                )

                if self.mismatched_node_couplings:
                    LOGGER.info(
                        "Found changing mismatched node couplings: %s",
                        [f"{n1.full_name} -> {n2.full_name}"
                         for n1, n2 in self.mismatched_node_couplings]
                    )
            else:
                self.trace_forest = TraceForest(self.model)
        else:
            # just make an empty TraceForest
            self.trace_forest = TraceForest(self.model)

    cdef initialise_model_data(self):
        self.model_data = ModelData()
        self.model_data.EPSILON0_C = self.model._EPSILON0_C
        self.model_data.UNIT_VACUUM = self.model._UNIT_VACUUM
        self.model_data.lambda0 = self.model.lambda0
        self.model_data.f0 = C_LIGHT / self.model_data.lambda0
        self.model_data.k0 = 2.0 * PI / self.model_data.lambda0
        self.model_data.homs_view = self.model.homs
        self.model_data.homs_ptr = <int*>&self.model_data.homs_view[0, 0]
        self.model_data.maxtem = np.max(self.model.homs[:, 0] + self.model.homs[:, 1])
        self.model_data.max_n = np.max(self.model.homs[:,0])
        self.model_data.max_m = np.max(self.model.homs[:,1])
        self.model_data.num_HOMs = len(self.model.homs)
        self.model_data.x_scale = self.model._x_scale
        if self.model.fsig.f.value is None:
            self.model_data.fsig = 0
        else:
            self.model_data.fsig = self.model.fsig.f.value
        self.model_data.zero_K00 = self.model.phase_config.ZERO_K00
        self.model_data.zero_tem00_gouy = self.model.phase_config.ZERO_TEM00_GOUY
        self.model_data.v2_transmission_phase = self.model.phase_config.V2_TRANSMISSION_PHASE

    cdef initialise_sim_config_data(self):
        # Nominal number of threads will be Nhoms / 10
        self.config_data.nthreads_homs = determine_nthreads_even(self.model_data.num_HOMs, 10)

        LOGGER.info("Using %d threads for HOM parallel loops.", self.config_data.nthreads_homs)

    def generate_carrier_frequencies(self):
        """Returns a list of Frequency objects that the model has requested"""
        from finesse.frequency import Frequency, generate_frequency_list
        frequencies_to_use = generate_frequency_list(self.model)

        carrier_frequencies = []

        LOGGER.info('Generating simulation with carrier frequencies %s', frequencies_to_use)
        for i, f in enumerate(frequencies_to_use):
            carrier_frequencies.append(
                Frequency(str(f.eval(keep_changing_symbols=True)), self, f, index=i)
            )

        fcnt = FrequencyContainer(carrier_frequencies)
        return fcnt

    def generate_signal_frequencies(self, nodes, FrequencyContainer carrier_optical_frequencies):
        """Generates the optical, mechanical, and electrical frequencies that should be
        modelled by the signal simulation.
        """
        signal_frequencies = [] # All optical frequencies are present at all nodes
        # elec and mech can have different frequencies on a per node basis
        mechanical_frequencies = {}
        electrical_frequencies = {}

        for i, f in enumerate(carrier_optical_frequencies.frequencies):
            fp = f.f + self.model.fsig.f.ref
            fm = f.f - self.model.fsig.f.ref
            signal_frequencies.append(
                Frequency(str(fp.eval(keep_changing_symbols=True)),
                            self, fp, index=2*i, audio_order=1,
                            audio=True, audio_carrier_index=i)
            )
            signal_frequencies.append(
                Frequency(str(fm.eval(keep_changing_symbols=True)),
                            self, fm, index=2*i+1, audio_order=-1,
                            audio=True, audio_carrier_index=i)
            )

        fcnt = FrequencyContainer(signal_frequencies, carrier_cnt=carrier_optical_frequencies)

        # Audio matrix frequencies are more complicated as they can have multiple frequencies
        # in mechanical and electrical, on a per-node basis...
        fsig = FrequencyContainer(
            (Frequency("fsig", self, self.model.fsig.f.ref, index=0), )
        )


        for node in nodes:
            if node.type == NodeType.OPTICAL:
                continue

            #-----------------------------------------------------------------------------------
            # Mechanical frequencies
            #-----------------------------------------------------------------------------------
            # By default mechanical frequencies just have a single frequency at the Model.fsig.f
            # However for more complicated systems we can have multiple frequencies.
            elif node.type == NodeType.MECHANICAL:
                fs = []
                freqs = node.frequencies
                if len(freqs) == 1 and freqs[0] == self.model.fsig.f.ref:
                    # Most components will just use a single fsig so reuse same object
                    # for efficient filling later
                    mechanical_frequencies[node] = fsig
                else:
                    for i, sym in enumerate(node.frequencies):
                        fs.append(
                            Frequency(
                                str(fm.eval(keep_changing_symbols=True)),
                                self,
                                sym,
                                index=i,
                            )
                        )
                    mechanical_frequencies[node] = FrequencyContainer(tuple(fs))

            #-----------------------------------------------------------------------------------
            # Electrical frequencies
            #-----------------------------------------------------------------------------------
            elif node.type == NodeType.ELECTRICAL:
                electrical_frequencies[node] = fsig
            else:
                raise ValueError("Unexpected")

        return fcnt, mechanical_frequencies, electrical_frequencies

    cpdef initialise_workspaces(self):
        from finesse.components import Connector, Cavity, Variable
        # TODO ddb - probably need to move away from lists as they aren't that fast to iterate
        # over. Maybe once we have all the lists filled we can covert them into some PyObject
        # memoryview

        self.workspaces = []
        self.cavity_workspaces = {}
        self.to_scatter_matrix_compute = []
        self.variable_workspaces = []
        self.gouy_phase_workspaces = []

        if self.is_modal and self.trace == NULL:
            raise Exception("Beam trace has not been set before workspaces are initialised")

        # Get any callbacks for the elements in the model
        # tell the element that we have now built the model and it
        # should do some initialisations for running simulations
        for el in self.model.elements.values():
            el._setup_changing_params()

            if isinstance(el, Connector):
                ws = el._get_workspace(self)
                if ws is None:
                    continue

                self.workspaces.append(ws) # store all workspaces here

                # if the connector scatters modes then initialise the
                # knm workspaces here and store the connector workspace
                # in to_scatter_matrix_compute for future use
                if self.is_modal and isinstance(ws, (KnmConnectorWorkspace)):
                    (<KnmConnectorWorkspace> ws).initialise_knm_workspaces()
                    self.to_scatter_matrix_compute.append(ws)

                if isinstance(ws, ConnectorWorkspace):
                    # Determine if we should be adding this workspace to any
                    # todo list for looping over later
                    if self.signal:
                        x = (
                            (ws.carrier, self.carrier),
                            (ws.signal, self.signal)
                        )
                    else:
                        x = ((ws.carrier, self.carrier),)

                    for info, mtx in x:
                        ws_store = mtx.workspaces
                        if info.callback_flag & ConnectorCallbacks.FILL_MATRIX:
                            ws_store.to_initial_fill.append(ws) # Initial fill all
                            if info.refill or mtx.forced_refill:
                                ws_store.to_refill.append(ws)

                        if info.callback_flag & ConnectorCallbacks.FILL_RHS:
                            ws_store.to_rhs_refill.append(ws)

                        if info.callback_flag & ConnectorCallbacks.FILL_NOISE:
                            try:
                                if any([x in self.signal.noise_sources for x in ws.owner.noises]):
                                    ws_store.to_noise_refill.append(ws)
                            except AttributeError:
                                # Component isn't a NoiseGenerator, but can still generate quantum
                                # noise
                                if NoiseType.QUANTUM in self.signal.noise_sources:
                                    ws_store.to_noise_refill.append(ws)

                        # Quantum noise is special, as all connectors can be a source if they have
                        # open inputs
                        if info.callback_flag & ConnectorCallbacks.FILL_INPUT_NOISE:
                            if NoiseType.QUANTUM in self.signal.noise_sources:
                                ws_store.to_noise_input_refill.append(ws)

                    if ws.fn_gouy_c is not None or ws.fn_gouy_py is not None:
                        self.gouy_phase_workspaces.append(ws)

                elif ws is not None:
                    # None means the component doesn't want anything
                    # to do with this simulation
                    raise Exception("Unexpected workspace type")
            elif isinstance(el, Variable):
                ws = el._get_workspace(self)
                if ws is not None:
                    self.variable_workspaces.append(ws)
            elif isinstance(el, Cavity):
                self.cavity_workspaces[el] = el._get_workspace(self)

        # Compile cy_exprs for changing symbolics, these are stored
        # in ElementWorkspace.chprm_expr which is used for fast evaluating
        # of the changing symbolic expressions
        for ws in self.workspaces:
            ws.compile_cy_exprs()
            # Also compile the changing ABCD matrix elements, these are
            # stored in the relevant cy_expr** field of the associated
            # workspace -> note that the cy_expr* element is NULL for non
            # changing elements
            if self.is_modal:
                ws.compile_abcd_cy_exprs()

        for vws in self.variable_workspaces:
            vws.compile_cy_exprs()

        self.carrier.workspaces.list_to_C()
        if self.signal:
            self.signal.workspaces.list_to_C()
        #print(self.name)
        #print("MATRIX")
        #print("carrier to_refill", len(self.carrier.workspaces.to_refill))
        #print("signal to_refill", len(self.signal.workspaces.to_refill))
        #print("to_rhs_refill", self.to_rhs_refill)

    cpdef initialise_noise_matrices(self):
        from finesse.detectors.general import NoiseDetector

        # Which noise types are we measuring?
        for el in self.model.detectors:
            if isinstance(el, NoiseDetector):
                if el.noise_type not in self.signal.noise_sources:
                    self.signal.noise_sources[el.noise_type] = []
                    self.signal.add_noise_matrix(el.noise_type)

    cpdef initialise_noise_sources(self):
        from finesse.components.general import NoiseGenerator

        for el in self.model.elements.values():
            if isinstance(el, NoiseGenerator):
                for _type, nodes in el.noises.items():
                    # Only consider noise sources for the types of noise we'll be measuring
                    if _type in self.signal.noise_sources:
                        self.signal.noise_sources[_type].append((el, nodes))

    cpdef initialise_noise_selection_vectors(self):
        for el in self.model.elements.values():
            if hasattr(el, "_requested_selection_vectors"):
                for name in el._requested_selection_vectors:
                    el._requested_selection_vectors[name] = self.signal._M.request_rhs_view()


cdef class MatrixSystemWorkspaces:
    def __init__(self):
        self.to_initial_fill = []
        self.to_refill = []
        self.to_rhs_refill = []
        self.to_noise_refill = []
        self.to_noise_input_refill = []
        self.noise_detectors = []

    def list_to_C(self):
        """Converts the python lists of workspaces into C Pyobject arrays for
        fast loop access.
        """
        self.num_to_refill = len(self.to_refill)
        self.ptr_to_refill = <PyObject**> malloc(self.num_to_refill * sizeof(PyObject*))
        if not self.ptr_to_refill:
            raise MemoryError()
        cdef int i
        for i in range(self.num_to_refill):
            self.ptr_to_refill[i] = <PyObject*>self.to_refill[i]

        self.num_to_rhs_refill = len(self.to_rhs_refill)
        self.ptr_to_rhs_refill = <PyObject**> malloc(self.num_to_rhs_refill * sizeof(PyObject*))
        if not self.ptr_to_rhs_refill:
            raise MemoryError()
        for i in range(self.num_to_rhs_refill):
            self.ptr_to_rhs_refill[i] = <PyObject*>self.to_rhs_refill[i]

        self.num_to_noise_refill = len(self.to_noise_refill)
        self.ptr_to_noise_refill = <PyObject**> malloc(self.num_to_noise_refill * sizeof(PyObject*))
        if not self.ptr_to_noise_refill:
            raise MemoryError()
        for i in range(self.num_to_noise_refill):
            self.ptr_to_noise_refill[i] = <PyObject*>self.to_noise_refill[i]

        self.num_to_noise_input_refill = len(self.to_noise_input_refill)
        self.ptr_to_noise_input_refill = <PyObject**> malloc(self.num_to_noise_input_refill * sizeof(PyObject*))
        if not self.ptr_to_noise_input_refill:
            raise MemoryError()
        for i in range(self.num_to_noise_input_refill):
            self.ptr_to_noise_input_refill[i] = <PyObject*>self.to_noise_input_refill[i]

    def detector_list_to_C(self):
        self.num_noise_detectors = len(self.noise_detectors)
        self.ptr_noise_detectors = <PyObject**> malloc(self.num_noise_detectors * sizeof(PyObject*))
        if not self.ptr_noise_detectors:
            raise MemoryError()
        for i in range(self.num_noise_detectors):
            self.ptr_noise_detectors[i] = <PyObject*>self.noise_detectors[i]

    def __dealloc__(self):
        if self.ptr_to_refill:
            free(self.ptr_to_refill)

        if self.ptr_to_rhs_refill:
            free(self.ptr_to_rhs_refill)

        if self.ptr_to_noise_refill:
            free(self.ptr_to_noise_refill)

        if self.ptr_to_noise_input_refill:
            free(self.ptr_to_noise_input_refill)

        if self.ptr_noise_detectors:
            free(self.ptr_noise_detectors)
